from __future__ import annotations

import logging
from collections import defaultdict
from typing import TYPE_CHECKING

from deptry.imports.extractors import NotebookImportExtractor
from deptry.rust import get_imports_from_py_files

if TYPE_CHECKING:
    from pathlib import Path

    from deptry.rust import Location as RustLocation

from deptry.imports.location import Location


def get_imported_modules_from_list_of_files(list_of_files: list[Path]) -> dict[str, list[Location]]:
    logging.info("Scanning %d %s...", len(list_of_files), "files" if len(list_of_files) > 1 else "file")

    py_files = [str(file) for file in list_of_files if file.suffix == ".py"]
    ipynb_files = [file for file in list_of_files if file.suffix == ".ipynb"]

    modules: dict[str, list[Location]] = defaultdict(list)

    # Process all .py files in parallel using Rust
    if py_files:
        rust_result = get_imports_from_py_files(py_files)
        for module, locations in convert_rust_locations_to_python_locations(rust_result).items():
            modules[module].extend(locations)

    # Process each .ipynb file individually
    for file in ipynb_files:
        for module, locations in get_imported_modules_from_ipynb_file(file).items():
            modules[module].extend(locations)

    logging.debug("All imported modules: %s\n", modules)

    return modules


def get_imported_modules_from_ipynb_file(path_to_file: Path) -> dict[str, list[Location]]:
    logging.debug("Scanning %s...", path_to_file)

    modules = NotebookImportExtractor(path_to_file).extract_imports()

    logging.debug("Found the following imports in %s: %s", path_to_file, modules)
    return modules


def convert_rust_locations_to_python_locations(
    imported_modules: dict[str, list[RustLocation]],
) -> dict[str, list[Location]]:
    converted_modules: dict[str, list[Location]] = {}
    for module, locations in imported_modules.items():
        converted_modules[module] = [Location.from_rust_location_object(loc) for loc in locations]
    return converted_modules



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING

from deptry.violations.base import ViolationsFinder
from deptry.violations.dep004_misplaced_dev.violation import DEP004MisplacedDevDependencyViolation

if TYPE_CHECKING:
    from deptry.module import Module
    from deptry.violations import Violation


@dataclass
class DEP004MisplacedDevDependenciesFinder(ViolationsFinder):
    """
    Given a list of imported modules and a list of project dependencies, determine which development dependencies
    should actually be regular dependencies.

    This is the case for any development dependency encountered, since files solely used for development purposes should be excluded from scanning.
    """

    def find(self) -> list[Violation]:
        """
        In this function, we use 'corresponding_package_name' instead of module.package, since it can happen that a
        development dependency is not installed, but it's still found to be used in the codebase, due to simple name
        matching. In that case, it's added under module.dev_top_levels. _get_package_name is added for these edge-cases.
        """
        logging.debug("\nScanning for incorrect development dependencies...")
        misplaced_dev_dependencies: list[Violation] = []

        for module_with_locations in self.imported_modules_with_locations:
            module = module_with_locations.module

            logging.debug("Scanning module %s...", module.name)
            corresponding_package_name = self._get_package_name(module)

            if corresponding_package_name and self._is_development_dependency(module, corresponding_package_name):
                for location in module_with_locations.locations:
                    misplaced_dev_dependencies.append(DEP004MisplacedDevDependencyViolation(module, location))

        return misplaced_dev_dependencies

    def _is_development_dependency(self, module: Module, corresponding_package_name: str) -> bool:
        # Module can be provided both by a regular and by a development dependency.
        # Only continue if module is ONLY provided by a dev dependency.
        if not module.is_provided_by_dev_dependency or module.is_provided_by_dependency:
            return False

        if module.name in self.ignored_modules:
            logging.debug(
                "Dependency '%s' found to be a misplaced development dependency, but ignoring.",
                corresponding_package_name,
            )
            return False

        logging.debug("Dependency '%s' marked as a misplaced development dependency.", corresponding_package_name)
        return True

    def _get_package_name(self, module: Module) -> str | None:
        if module.package:
            return module.package
        if module.dev_top_levels:
            if len(module.dev_top_levels) > 1:
                logging.debug(
                    "Module %s is found in the top-level module names of multiple development dependencies. Skipping.",
                    module.name,
                )
            elif len(module.dev_top_levels) == 0:
                logging.debug(
                    "Module %s has no metadata and it is not found in any top-level module names. Skipping.",
                    module.name,
                )
            else:
                return module.dev_top_levels[0]
        return None



from __future__ import annotations

from pathlib import Path

import pytest
from pathspec import PathSpec
from pathspec.patterns.gitwildmatch import GitWildMatchPattern

from deptry.python_file_finder import PythonFileFinder
from tests.utils import create_files, run_within_dir


def test_simple(tmp_path: Path) -> None:
    with run_within_dir(tmp_path):
        create_files([
            Path("dir/subdir/file1.py"),
            Path("dir/subdir/file2.py"),
            Path("dir/subdir/file3.py"),
            Path("other_dir/subdir/file1.py"),
            Path("other_dir/subdir/file2.py"),
        ])

        files = PythonFileFinder(
            exclude=(".venv",), extend_exclude=("other_dir",), using_default_exclude=False
        ).get_all_python_files_in((Path(),))

        assert sorted(files) == [
            Path("dir/subdir/file1.py"),
            Path("dir/subdir/file2.py"),
            Path("dir/subdir/file3.py"),
        ]


def test_only_matches_start(tmp_path: Path) -> None:
    """
    Test that adding 'subdir' as exclude argument does not also exclude dir/subdir.
    """
    with run_within_dir(tmp_path):
        create_files([
            Path("dir/subdir/file1.py"),
            Path("dir/subdir/file2.py"),
            Path("dir/subdir/file3.py"),
            Path("subdir/file1.py"),
            Path("subdir/file2.py"),
        ])

        files = PythonFileFinder(
            exclude=("subdir",), extend_exclude=(), using_default_exclude=False
        ).get_all_python_files_in((Path(),))

        assert sorted(files) == [
            Path("dir/subdir/file1.py"),
            Path("dir/subdir/file2.py"),
            Path("dir/subdir/file3.py"),
        ]


@pytest.mark.parametrize(
    ("ignore_notebooks", "expected"),
    [
        (
            False,
            [Path("dir/subdir/file1.ipynb")],
        ),
        (
            True,
            [],
        ),
    ],
)
def test_matches_ipynb(ignore_notebooks: bool, expected: list[Path], tmp_path: Path) -> None:
    with run_within_dir(tmp_path):
        create_files([Path("dir/subdir/file1.ipynb")])

        files = PythonFileFinder(
            exclude=(), extend_exclude=(), using_default_exclude=False, ignore_notebooks=ignore_notebooks
        ).get_all_python_files_in((Path(),))

        assert sorted(files) == expected


@pytest.mark.parametrize(
    ("exclude", "expected"),
    [
        (
            (".*file1",),
            [
                Path(".cache/file2.py"),
                Path("dir/subdir/file2.py"),
                Path("dir/subdir/file3.py"),
                Path("other_dir/subdir/file2.py"),
            ],
        ),
        (
            (".cache|other.*subdir",),
            [
                Path("dir/subdir/file1.py"),
                Path("dir/subdir/file2.py"),
                Path("dir/subdir/file3.py"),
            ],
        ),
        (
            (".*/subdir/",),
            [
                Path(".cache/file1.py"),
                Path(".cache/file2.py"),
            ],
        ),
    ],
)
def test_regex_argument(exclude: tuple[str], expected: list[Path], tmp_path: Path) -> None:
    with run_within_dir(tmp_path):
        create_files([
            Path(".cache/file1.py"),
            Path(".cache/file2.py"),
            Path("dir/subdir/file1.py"),
            Path("dir/subdir/file2.py"),
            Path("dir/subdir/file3.py"),
            Path("other_dir/subdir/file1.py"),
            Path("other_dir/subdir/file2.py"),
        ])

        files = PythonFileFinder(
            exclude=exclude, extend_exclude=(), using_default_exclude=False
        ).get_all_python_files_in((Path(),))

        assert sorted(files) == expected


@pytest.mark.parametrize(
    ("exclude", "expected"),
    [
        (
            (".*file1",),
            [
                Path("dir/subdir/file2.py"),
                Path("dir/subdir/file3.py"),
                Path("other_dir/subdir/file2.py"),
            ],
        ),
        (
            (".*file1|.*file2",),
            [
                Path("dir/subdir/file3.py"),
            ],
        ),
        (
            (".*/subdir/",),
            [],
        ),
    ],
)
def test_multiple_source_directories(exclude: tuple[str], expected: list[Path], tmp_path: Path) -> None:
    with run_within_dir(tmp_path):
        create_files([
            Path("dir/subdir/file1.py"),
            Path("dir/subdir/file2.py"),
            Path("dir/subdir/file3.py"),
            Path("other_dir/subdir/file1.py"),
            Path("other_dir/subdir/file2.py"),
            Path("another_dir/subdir/file1.py"),
        ])

        files = PythonFileFinder(
            exclude=exclude, extend_exclude=(), using_default_exclude=False
        ).get_all_python_files_in((Path("dir"), Path("other_dir")))

        assert sorted(files) == expected


def test_duplicates_are_removed(tmp_path: Path) -> None:
    with run_within_dir(tmp_path):
        create_files([Path("dir/subdir/file1.py")])

        files = PythonFileFinder(exclude=(), extend_exclude=(), using_default_exclude=False).get_all_python_files_in((
            Path(),
            Path(),
        ))

        assert sorted(files) == [Path("dir/subdir/file1.py")]


def test__generate_gitignore_pathspec_with_non_default_exclude(tmp_path: Path) -> None:
    gitignore_pathspec = PythonFileFinder(
        exclude=(), extend_exclude=(), using_default_exclude=False
    )._generate_gitignore_pathspec(Path())

    assert gitignore_pathspec is None


def test__generate_gitignore_pathspec_with_non_existing_gitignore(tmp_path: Path) -> None:
    with run_within_dir(tmp_path):
        gitignore_pathspec = PythonFileFinder(
            exclude=(), extend_exclude=(), using_default_exclude=True
        )._generate_gitignore_pathspec(Path())

        assert gitignore_pathspec is None


def test__generate_gitignore_pathspec_with_existing_gitignore(tmp_path: Path) -> None:
    with run_within_dir(tmp_path):
        with Path(".gitignore").open("w") as gitignore:
            gitignore.write("foo.py\nbar/")

        gitignore_pathspec = PythonFileFinder(
            exclude=(), extend_exclude=(), using_default_exclude=True
        )._generate_gitignore_pathspec(Path())

        assert isinstance(gitignore_pathspec, PathSpec)
        assert gitignore_pathspec.patterns == [
            GitWildMatchPattern("foo.py"),
            GitWildMatchPattern("bar/"),
        ]



from __future__ import annotations

import json
import os
import shlex
import shutil
import subprocess
import sys
import venv
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Generator

from deptry.reporters.text import COLORS


@dataclass
class _BaseVenvFactory:
    venvs_directory: Path

    @contextmanager
    def venv(self, project: str, setup_commands: list[str]) -> Generator[VirtualEnvironment, None, None]:
        venv_path = self.venvs_directory / project

        old_cwd = Path.cwd()

        virtual_env = VirtualEnvironment(venv_path, project)
        virtual_env.setup(setup_commands, old_cwd)

        os.chdir(venv_path / "project")

        try:
            yield virtual_env
        finally:
            os.chdir(old_cwd)


class PDMVenvFactory(_BaseVenvFactory):
    @contextmanager
    def __call__(self, project: str) -> Generator[VirtualEnvironment, None, None]:
        with self.venv(project, ["pip install pdm", "pdm install --no-self"]) as virtual_env:
            yield virtual_env


@dataclass
class PoetryVenvFactory(_BaseVenvFactory):
    @contextmanager
    def __call__(self, project: str) -> Generator[VirtualEnvironment, None, None]:
        with self.venv(project, ["pip install poetry", "poetry install --no-root"]) as virtual_env:
            yield virtual_env


class PipVenvFactory(_BaseVenvFactory):
    @contextmanager
    def __call__(
        self, project: str, install_command: str = "pip install ."
    ) -> Generator[VirtualEnvironment, None, None]:
        with self.venv(project, [install_command]) as virtual_env:
            yield virtual_env


@dataclass
class VirtualEnvironment:
    project_path: Path
    project: str
    python_executable: Path | None = None

    def __post_init__(self) -> None:
        self.python_executable = self.project_path / ("Scripts/python.exe" if sys.platform == "win32" else "bin/python")

    def setup(self, setup_commands: list[str], deptry_directory: Path) -> None:
        if self.project_path.exists():
            return

        virtual_env = venv.EnvBuilder(with_pip=True)
        virtual_env.create(self.project_path)

        shutil.copytree(deptry_directory / "tests/data" / self.project, self.project_path / "project")

        for setup_command in [f"pip install {deptry_directory}", *setup_commands]:
            self.run(setup_command, check=True, cwd=self.project_path / "project")

    def run(self, command: str, check: bool = False, cwd: Path | None = None) -> subprocess.CompletedProcess[str]:
        return subprocess.run(
            shlex.split(f"{self.python_executable} -m {command}", posix=sys.platform != "win32"),
            env={**os.environ, "VIRTUAL_ENV": str(self.project_path)},
            capture_output=True,
            text=True,
            check=check,
            cwd=cwd,
        )


@contextmanager
def run_within_dir(path: Path) -> Generator[None, None, None]:
    """
    Utility function to run some code within a directory, and change back to the current directory afterwards.

    Example usage:

    ```
    with run_within_dir(directory):
        some_code()
    ```

    """
    oldpwd = Path.cwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(oldpwd)


def get_issues_report(path: Path) -> list[dict[str, Any]]:
    with path.open() as file:
        report: list[dict[str, Any]] = json.load(file)

    return report


def create_files(paths: list[Path]) -> None:
    """
    Takes as input an argument paths, which is a list of dicts. Each dict should have two keys;
    'dir' to denote a directory and 'file' to denote the file name. This function creates all files
    within their corresponding directories.
    """
    for path in paths:
        path.parent.mkdir(parents=True, exist_ok=True)
        path.touch()


def stylize(text: str, **kwargs: Any) -> str:
    return text.format(**kwargs, **COLORS)



from __future__ import annotations

from pathlib import Path

from deptry.dependency_getter.pdm import PDMDependencyGetter
from tests.utils import run_within_dir


def test_dependency_getter(tmp_path: Path) -> None:
    fake_pyproject_toml = """\
[project]
# PEP 621 project metadata
# See https://www.python.org/dev/peps/pep-0621/
dependencies = [
    "qux",
    "bar>=20.9",
    "optional-foo[option]>=0.12.11",
    "conditional-bar>=1.1.0; python_version < 3.11",
    "fox-python",  # top level module is called "fox"
]
"""

    with run_within_dir(tmp_path):
        with Path("pyproject.toml").open("w") as f:
            f.write(fake_pyproject_toml)

        getter = PDMDependencyGetter(
            config=Path("pyproject.toml"),
            package_module_name_map={"fox-python": ("fox",)},
        )
        dependencies = getter.get().dependencies

        assert len(dependencies) == 5

        assert dependencies[0].name == "qux"
        assert not dependencies[0].is_conditional
        assert not dependencies[0].is_optional
        assert "qux" in dependencies[0].top_levels

        assert dependencies[1].name == "bar"
        assert not dependencies[1].is_conditional
        assert not dependencies[1].is_optional
        assert "bar" in dependencies[1].top_levels

        assert dependencies[2].name == "optional-foo"
        assert not dependencies[2].is_conditional
        assert dependencies[2].is_optional
        assert "optional_foo" in dependencies[2].top_levels

        assert dependencies[3].name == "conditional-bar"
        assert dependencies[3].is_conditional
        assert not dependencies[3].is_optional
        assert "conditional_bar" in dependencies[3].top_levels

        assert dependencies[4].name == "fox-python"
        assert not dependencies[4].is_conditional
        assert not dependencies[4].is_optional
        assert "fox" in dependencies[4].top_levels


def test_dev_dependency_getter(tmp_path: Path) -> None:
    fake_pyproject_toml = """\
[project]
# PEP 621 project metadata
# See https://www.python.org/dev/peps/pep-0621/
dependencies = [
    "qux",
    "bar>=20.9",
    "optional-foo[option]>=0.12.11",
    "conditional-bar>=1.1.0; python_version < 3.11",
]
[tool.pdm.dev-dependencies]
test = [
    "qux",
    "bar; python_version < 3.11"
    ]
tox = [
    "foo-bar",
]
"""

    with run_within_dir(tmp_path):
        with Path("pyproject.toml").open("w") as f:
            f.write(fake_pyproject_toml)

        dev_dependencies = PDMDependencyGetter(Path("pyproject.toml")).get().dev_dependencies

        assert len(dev_dependencies) == 3

        assert dev_dependencies[0].name == "qux"
        assert not dev_dependencies[0].is_conditional
        assert not dev_dependencies[0].is_optional
        assert "qux" in dev_dependencies[0].top_levels

        assert dev_dependencies[1].name == "bar"
        assert dev_dependencies[1].is_conditional
        assert not dev_dependencies[1].is_optional
        assert "bar" in dev_dependencies[1].top_levels

        assert dev_dependencies[2].name == "foo-bar"
        assert not dev_dependencies[2].is_conditional
        assert not dev_dependencies[2].is_optional
        assert "foo_bar" in dev_dependencies[2].top_levels



from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from deptry.dependency import Dependency
from deptry.imports.location import Location
from deptry.module import Module
from deptry.reporters import TextReporter
from deptry.reporters.text import COLORS, COLORS_NOOP
from deptry.violations import (
    DEP001MissingDependencyViolation,
    DEP002UnusedDependencyViolation,
    DEP003TransitiveDependencyViolation,
    DEP004MisplacedDevDependencyViolation,
)
from tests.utils import stylize

if TYPE_CHECKING:
    from _pytest.logging import LogCaptureFixture


def test_logging_number_multiple(caplog: LogCaptureFixture) -> None:
    with caplog.at_level(logging.INFO):
        violations = [
            DEP001MissingDependencyViolation(Module("foo", package="foo_package"), Location(Path("foo.py"), 1, 2)),
            DEP002UnusedDependencyViolation(
                Dependency("foo", Path("pyproject.toml")), Location(Path("pyproject.toml"))
            ),
            DEP003TransitiveDependencyViolation(
                Module("foo", package="foo_package"), Location(Path("foo/bar.py"), 1, 2)
            ),
            DEP004MisplacedDevDependencyViolation(Module("foo", package="foo_package"), Location(Path("foo.py"), 1, 2)),
        ]
        TextReporter(violations).report()

    assert caplog.messages == [
        (
            "Assuming the corresponding module name of package 'foo' is 'foo'. Install the package or configure a"
            " package_module_name_map entry to override this behaviour."
        ),
        "",
        stylize(
            "{BOLD}{file}{RESET}{CYAN}:{RESET}1{CYAN}:{RESET}2{CYAN}:{RESET} {BOLD}{RED}DEP001{RESET} 'foo'"
            " imported but missing from the dependency definitions",
            file=Path("foo.py"),
        ),
        stylize(
            "{BOLD}{file}{RESET}{CYAN}:{RESET} {BOLD}{RED}DEP002{RESET} 'foo' defined as a dependency but not used"
            " in the codebase",
            file=Path("pyproject.toml"),
        ),
        stylize(
            "{BOLD}{file}{RESET}{CYAN}:{RESET}1{CYAN}:{RESET}2{CYAN}:{RESET} {BOLD}{RED}DEP003{RESET} 'foo_package'"
            " imported but it is a transitive dependency",
            file=Path("foo/bar.py"),
        ),
        stylize(
            "{BOLD}{file}{RESET}{CYAN}:{RESET}1{CYAN}:{RESET}2{CYAN}:{RESET} {BOLD}{RED}DEP004{RESET} 'foo'"
            " imported but declared as a dev dependency",
            file=Path("foo.py"),
        ),
        stylize("{BOLD}{RED}Found 4 dependency issues.{RESET}"),
        "\nFor more information, see the documentation: https://fpgmaas.github.io/deptry/",
    ]


def test_logging_number_single(caplog: LogCaptureFixture) -> None:
    with caplog.at_level(logging.INFO):
        TextReporter([
            DEP001MissingDependencyViolation(Module("foo", package="foo_package"), Location(Path("foo.py"), 1, 2))
        ]).report()

    assert caplog.messages == [
        "",
        stylize(
            "{BOLD}{file}{RESET}{CYAN}:{RESET}1{CYAN}:{RESET}2{CYAN}:{RESET} {BOLD}{RED}DEP001{RESET} 'foo'"
            " imported but missing from the dependency definitions",
            file=Path("foo.py"),
        ),
        stylize("{BOLD}{RED}Found 1 dependency issue.{RESET}"),
        "\nFor more information, see the documentation: https://fpgmaas.github.io/deptry/",
    ]


def test_logging_number_none(caplog: LogCaptureFixture) -> None:
    with caplog.at_level(logging.INFO):
        TextReporter([]).report()

    assert caplog.messages == [
        "",
        stylize("{BOLD}{GREEN}Success! No dependency issues found.{RESET}"),
    ]


def test_logging_no_ansi(caplog: LogCaptureFixture) -> None:
    with caplog.at_level(logging.INFO):
        violations = [
            DEP001MissingDependencyViolation(Module("foo", package="foo_package"), Location(Path("foo.py"), 1, 2)),
            DEP002UnusedDependencyViolation(
                Dependency("foo", Path("pyproject.toml")), Location(Path("pyproject.toml"))
            ),
            DEP003TransitiveDependencyViolation(
                Module("foo", package="foo_package"), Location(Path("foo/bar.py"), 1, 2)
            ),
            DEP004MisplacedDevDependencyViolation(Module("foo", package="foo_package"), Location(Path("foo.py"), 1, 2)),
        ]
        TextReporter(violations, use_ansi=False).report()

    assert caplog.messages == [
        (
            "Assuming the corresponding module name of package 'foo' is 'foo'. Install the package or configure a"
            " package_module_name_map entry to override this behaviour."
        ),
        "",
        f"{Path('foo.py')}:1:2: DEP001 'foo' imported but missing from the dependency definitions",
        f"{Path('pyproject.toml')}: DEP002 'foo' defined as a dependency but not used in the codebase",
        f"{Path('foo/bar.py')}:1:2: DEP003 'foo_package' imported but it is a transitive dependency",
        f"{Path('foo.py')}:1:2: DEP004 'foo' imported but declared as a dev dependency",
        "Found 4 dependency issues.",
        "\nFor more information, see the documentation: https://fpgmaas.github.io/deptry/",
    ]


@pytest.mark.parametrize(
    ("use_ansi", "expected"),
    [
        (True, COLORS),
        (False, COLORS_NOOP),
    ],
)
def test__get_colors(use_ansi: bool, expected: dict[str, str]) -> None:
    assert TextReporter([], use_ansi)._get_colors() == expected



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from deptry.violations import Violation


@dataclass
class Reporter(ABC):
    """Base class for all violation reporters."""

    violations: list[Violation]

    @abstractmethod
    def report(self) -> None:
        raise NotImplementedError()



from __future__ import annotations

from pathlib import Path

from deptry.dependency import Dependency
from deptry.imports.location import Location
from deptry.module import Module, ModuleLocations
from deptry.violations import DEP004MisplacedDevDependencyViolation
from deptry.violations.dep004_misplaced_dev.finder import DEP004MisplacedDevDependenciesFinder


def test_simple() -> None:
    dependencies = [Dependency("bar", Path("pyproject.toml"))]

    module_foo_locations = [Location(Path("foo.py"), 1, 2), Location(Path("bar.py"), 3, 4)]
    module_foo = Module("foo", dev_top_levels=["foo"], is_provided_by_dev_dependency=True)

    modules_locations = [ModuleLocations(module_foo, module_foo_locations)]

    assert DEP004MisplacedDevDependenciesFinder(modules_locations, dependencies).find() == [
        DEP004MisplacedDevDependencyViolation(module_foo, location) for location in module_foo_locations
    ]


def test_regular_and_dev_dependency() -> None:
    """
    If a dependency is both a regular and a development dependency, no 'misplaced dev dependency' violation
    should be detected if it is used in the code base.
    """

    dependencies = [Dependency("foo", Path("pyproject.toml"))]

    module_foo_locations = [Location(Path("foo.py"), 1, 2)]
    module_foo = Module(
        "foo", dev_top_levels=["foo"], is_provided_by_dev_dependency=True, is_provided_by_dependency=True
    )

    modules_locations = [ModuleLocations(module_foo, module_foo_locations)]

    assert not DEP004MisplacedDevDependenciesFinder(modules_locations, dependencies).find()



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING

from deptry.dependency_getter.base import DependenciesExtract
from deptry.dependency_getter.pep_621 import PEP621DependencyGetter
from deptry.utils import load_pyproject_toml

if TYPE_CHECKING:
    from deptry.dependency import Dependency


@dataclass
class PDMDependencyGetter(PEP621DependencyGetter):
    """
    Class to get dependencies that are specified according to PEP 621 from a `pyproject.toml` file for a project that uses PDM for its dependency management.
    """

    def get(self) -> DependenciesExtract:
        pep_621_dependencies_extract = super().get()

        dev_dependencies = self._get_pdm_dev_dependencies()
        self._log_dependencies(dev_dependencies, is_dev=True)

        return DependenciesExtract(pep_621_dependencies_extract.dependencies, dev_dependencies)

    def _get_pdm_dev_dependencies(self) -> list[Dependency]:
        """
        Try to get development dependencies from pyproject.toml, which with PDM are specified as:

        [tool.pdm.dev-dependencies]
        test = [
            "pytest",
            "pytest-cov",
        ]
        tox = [
            "tox",
            "tox-pdm>=0.5",
        ]
        """
        pyproject_data = load_pyproject_toml(self.config)

        dev_dependency_strings: list[str] = []
        try:
            dev_dependencies_dict: dict[str, str] = pyproject_data["tool"]["pdm"]["dev-dependencies"]
            for deps in dev_dependencies_dict.values():
                dev_dependency_strings += deps
        except KeyError:
            logging.debug("No section [tool.pdm.dev-dependencies] found in pyproject.toml")

        return self._extract_pep_508_dependencies(dev_dependency_strings, self.package_module_name_map)



from __future__ import annotations

import json
import logging
import re
import uuid
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from deptry.imports.extract import get_imported_modules_from_ipynb_file, get_imported_modules_from_list_of_files
from deptry.imports.location import Location
from tests.utils import run_within_dir

if TYPE_CHECKING:
    from _pytest.logging import LogCaptureFixture


def test_import_parser_py() -> None:
    some_imports_path = Path("tests/data/some_imports.py")

    assert get_imported_modules_from_list_of_files([some_imports_path]) == {
        "barfoo": [Location(some_imports_path, 20, 8)],
        "baz": [Location(some_imports_path, 16, 5)],
        "click": [Location(some_imports_path, 24, 12)],
        "foobar": [Location(some_imports_path, 18, 12)],
        "httpx": [Location(some_imports_path, 14, 12)],
        "module_in_class": [Location(some_imports_path, 35, 16)],
        "module_in_func": [Location(some_imports_path, 30, 12)],
        "not_click": [Location(some_imports_path, 26, 12)],
        "numpy": [
            Location(some_imports_path, 5, 8),
            Location(some_imports_path, 7, 1),
        ],
        "os": [Location(some_imports_path, 1, 1)],
        "pandas": [Location(some_imports_path, 6, 8)],
        "pathlib": [Location(some_imports_path, 2, 1)],
        "randomizer": [Location(some_imports_path, 21, 1)],
        "typing": [Location(some_imports_path, 3, 1)],
    }


def test_import_parser_ipynb() -> None:
    notebook_path = Path("tests/data/example_project/src/notebook.ipynb")

    assert get_imported_modules_from_ipynb_file(notebook_path) == {
        "click": [Location(notebook_path, 1, 0)],
        "toml": [Location(notebook_path, 5, 0)],
        "urllib3": [Location(notebook_path, 3, 0)],
    }


@pytest.mark.parametrize(
    ("file_content", "encoding"),
    [
        (
            "# -*- encoding: utf-8 -*-\nimport foo\nprint('嘉大')",
            "utf-8",
        ),
        (
            "# -*- encoding: iso-8859-15 -*-\nimport foo\nprint('Æ	Ç')",
            "iso-8859-15",
        ),
        (
            "# -*- encoding: utf-16 -*-\nimport foo\nprint('嘉大')",
            "utf-16",
        ),
        (
            "\nimport foo\nmy_string = '🐺'",
            "utf-16",
        ),
    ],
)
def test_import_parser_file_encodings(file_content: str, encoding: str | None, tmp_path: Path) -> None:
    random_file = Path(f"file_{uuid.uuid4()}.py")

    with run_within_dir(tmp_path):
        with random_file.open("w", encoding=encoding) as f:
            f.write(file_content)

        assert get_imported_modules_from_list_of_files([random_file]) == {"foo": [Location(random_file, 2, 8)]}


@pytest.mark.parametrize(
    ("code_cell_content", "encoding"),
    [
        (
            ["import foo", "print('嘉大')"],
            "utf-8",
        ),
        (
            ["import foo", "print('Æ	Ç')"],
            "iso-8859-15",
        ),
        (
            ["import foo", "print('嘉大')"],
            "utf-16",
        ),
        (
            ["import foo", "my_string = '🐺'"],
            None,
        ),
    ],
)
def test_import_parser_file_encodings_ipynb(code_cell_content: list[str], encoding: str | None, tmp_path: Path) -> None:
    random_file = Path(f"file_{uuid.uuid4()}.ipynb")

    with run_within_dir(tmp_path):
        with random_file.open("w", encoding=encoding) as f:
            file_content = {
                "cells": [
                    {
                        "cell_type": "code",
                        "metadata": {},
                        "source": code_cell_content,
                    }
                ]
            }
            f.write(json.dumps(file_content))

        assert get_imported_modules_from_list_of_files([random_file]) == {"foo": [Location(random_file, 1, 0)]}


def test_import_parser_file_encodings_warning(tmp_path: Path, caplog: LogCaptureFixture) -> None:
    file_path = Path("file1.py")

    with run_within_dir(tmp_path):
        # The characters below are represented differently in ISO-8859-1 and UTF-8, so this should raise an error.
        with file_path.open("w", encoding="ISO-8859-1") as f:
            f.write("# -*- coding: utf-8 -*-\nprint('ÆØÅ')")

        with caplog.at_level(logging.WARNING):
            assert get_imported_modules_from_list_of_files([file_path]) == {}

        # //TODO logging from Rust still includes it's own warning and file + line number. Can we get rid of that?
        pattern = re.compile(
            r"WARNING  deptry.imports:imports.rs:\d+ Warning: File file1.py could not be read. Skipping...\n"
        )
        assert pattern.search(caplog.text) is not None



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING

from deptry.violations.base import ViolationsFinder
from deptry.violations.dep001_missing.violation import DEP001MissingDependencyViolation

if TYPE_CHECKING:
    from deptry.module import Module
    from deptry.violations.base import Violation


@dataclass
class DEP001MissingDependenciesFinder(ViolationsFinder):
    """
    Given a list of imported modules and a list of project dependencies, determine which ones are missing.
    """

    def find(self) -> list[Violation]:
        logging.debug("\nScanning for missing dependencies...")
        missing_dependencies: list[Violation] = []

        for module_with_locations in self.imported_modules_with_locations:
            module = module_with_locations.module

            logging.debug("Scanning module %s...", module.name)

            if self._is_missing(module):
                for location in module_with_locations.locations:
                    missing_dependencies.append(DEP001MissingDependencyViolation(module, location))

        return missing_dependencies

    def _is_missing(self, module: Module) -> bool:
        if any([
            module.package is not None,
            module.is_provided_by_dependency,
            module.is_provided_by_dev_dependency,
            module.local_module,
        ]):
            return False

        if module.name in self.ignored_modules:
            logging.debug("Identified module '%s' as a missing dependency, but ignoring.", module.name)
            return False

        logging.debug("No package found to import module '%s' from. Marked as a missing dependency.", module.name)
        return True



from __future__ import annotations

from pathlib import Path

from deptry.dependency_getter.poetry import PoetryDependencyGetter
from tests.utils import run_within_dir


def test_dependency_getter(tmp_path: Path) -> None:
    fake_pyproject_toml = """
[tool.poetry.dependencies]
python = ">=3.7,<4.0"
bar =  { version = ">=2.5.1,<4.0.0", python = ">3.7" }
foo-bar =  { version = ">=2.5.1,<4.0.0", optional = true, python = ">3.7" }
fox-python = "*"  # top level module is called "fox"

[tool.poetry.dev-dependencies]
toml = "^0.10.2"
qux =  { version = ">=2.5.1,<4.0.0", optional = true }

[tool.poetry.group.lint.dependencies]
black = "^22.6.0"
mypy = "^1.3.0"

[tool.poetry.group.test.dependencies]
pytest = "^7.3.0"
pytest-cov = "^4.0.0"
"""

    with run_within_dir(tmp_path):
        with Path("pyproject.toml").open("w") as f:
            f.write(fake_pyproject_toml)

        getter = PoetryDependencyGetter(
            config=Path("pyproject.toml"),
            package_module_name_map={"fox-python": ("fox",)},
        )
        dependencies_extract = getter.get()
        dependencies = dependencies_extract.dependencies
        dev_dependencies = dependencies_extract.dev_dependencies

        assert len(dependencies) == 3
        assert len(dev_dependencies) == 6

        assert dependencies[0].name == "bar"
        assert dependencies[0].is_conditional
        assert not dependencies[0].is_optional
        assert "bar" in dependencies[0].top_levels

        assert dependencies[1].name == "foo-bar"
        assert dependencies[1].is_conditional
        assert dependencies[1].is_optional
        assert "foo_bar" in dependencies[1].top_levels

        assert dependencies[2].name == "fox-python"
        assert not dependencies[2].is_conditional
        assert not dependencies[2].is_optional
        assert "fox" in dependencies[2].top_levels

        assert dev_dependencies[0].name == "toml"
        assert not dev_dependencies[0].is_conditional
        assert not dev_dependencies[0].is_optional
        assert "toml" in dev_dependencies[0].top_levels

        assert dev_dependencies[1].name == "qux"
        assert not dev_dependencies[1].is_conditional
        assert dev_dependencies[1].is_optional
        assert "qux" in dev_dependencies[1].top_levels

        assert dev_dependencies[2].name == "black"
        assert not dev_dependencies[2].is_conditional
        assert not dev_dependencies[2].is_optional
        assert "black" in dev_dependencies[2].top_levels

        assert dev_dependencies[3].name == "mypy"
        assert not dev_dependencies[3].is_conditional
        assert not dev_dependencies[3].is_optional
        assert "mypy" in dev_dependencies[3].top_levels

        assert dev_dependencies[4].name == "pytest"
        assert not dev_dependencies[4].is_conditional
        assert not dev_dependencies[4].is_optional
        assert "pytest" in dev_dependencies[4].top_levels

        assert dev_dependencies[5].name == "pytest-cov"
        assert not dev_dependencies[5].is_conditional
        assert not dev_dependencies[5].is_optional
        assert "pytest_cov" in dev_dependencies[5].top_levels



from __future__ import annotations

import sys
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


from deptry.exceptions import PyprojectFileNotFoundError

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib


def load_pyproject_toml(config: Path) -> dict[str, Any]:
    try:
        with config.open("rb") as pyproject_file:
            return tomllib.load(pyproject_file)
    except FileNotFoundError:
        raise PyprojectFileNotFoundError(Path.cwd()) from None



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from deptry.rust import Location as RustLocation


@dataclass(frozen=True)
class Location:
    file: Path
    line: int | None = None
    column: int | None = None

    @classmethod
    def from_rust_location_object(cls, location: RustLocation) -> Location:
        return cls(file=Path(location.file), line=location.line, column=location.column)



from __future__ import annotations

import re
from pathlib import Path

import pytest

from deptry.dependency_specification_detector import DependencyManagementFormat, DependencySpecificationDetector
from deptry.exceptions import DependencySpecificationNotFoundError
from tests.utils import run_within_dir


def test_poetry(tmp_path: Path) -> None:
    with run_within_dir(tmp_path):
        pyproject_toml_path = Path("pyproject.toml")

        with pyproject_toml_path.open("w") as f:
            f.write('[tool.poetry.dependencies]\nfake = "10"')

        spec = DependencySpecificationDetector(pyproject_toml_path).detect()
        assert spec == DependencyManagementFormat.POETRY


def test_requirements_txt(tmp_path: Path) -> None:
    with run_within_dir(tmp_path):
        with Path("requirements.txt").open("w") as f:
            f.write('foo >= "1.0"')

        spec = DependencySpecificationDetector(Path("pyproject.toml")).detect()
        assert spec == DependencyManagementFormat.REQUIREMENTS_TXT


def test_pdm_with_dev_dependencies(tmp_path: Path) -> None:
    with run_within_dir(tmp_path):
        pyproject_toml_path = Path("pyproject.toml")

        with pyproject_toml_path.open("w") as f:
            f.write(
                '[project]\ndependencies=["foo"]\n[tool.pdm]\nversion = {source ='
                ' "scm"}\n[tool.pdm.dev-dependencies]\ngroup=["bar"]'
            )

        spec = DependencySpecificationDetector(pyproject_toml_path).detect()
        assert spec == DependencyManagementFormat.PDM


def test_pdm_without_dev_dependencies(tmp_path: Path) -> None:
    with run_within_dir(tmp_path):
        pyproject_toml_path = Path("pyproject.toml")

        with pyproject_toml_path.open("w") as f:
            f.write('[project]\ndependencies=["foo"]\n[tool.pdm]\nversion = {source = "scm"}')

        spec = DependencySpecificationDetector(pyproject_toml_path).detect()
        assert spec == DependencyManagementFormat.PEP_621


def test_pep_621(tmp_path: Path) -> None:
    with run_within_dir(tmp_path):
        pyproject_toml_path = Path("pyproject.toml")

        with pyproject_toml_path.open("w") as f:
            f.write('[project]\ndependencies=["foo"]')

        spec = DependencySpecificationDetector(pyproject_toml_path).detect()
        assert spec == DependencyManagementFormat.PEP_621


def test_both(tmp_path: Path) -> None:
    """
    If both are found, result should be 'poetry'
    """

    with run_within_dir(tmp_path):
        pyproject_toml_path = Path("pyproject.toml")

        with pyproject_toml_path.open("w") as f:
            f.write('[tool.poetry.dependencies]\nfake = "10"')

        with Path("requirements.txt").open("w") as f:
            f.write('foo >= "1.0"')

        spec = DependencySpecificationDetector(pyproject_toml_path).detect()
        assert spec == DependencyManagementFormat.POETRY


def test_requirements_txt_with_argument(tmp_path: Path) -> None:
    with run_within_dir(tmp_path):
        with Path("req.txt").open("w") as f:
            f.write('foo >= "1.0"')

        spec = DependencySpecificationDetector(Path("pyproject.toml"), requirements_txt=("req.txt",)).detect()
        assert spec == DependencyManagementFormat.REQUIREMENTS_TXT


def test_requirements_txt_with_argument_not_root_directory(tmp_path: Path) -> None:
    with run_within_dir(tmp_path):
        Path("req").mkdir()

        with Path("req/req.txt").open("w") as f:
            f.write('foo >= "1.0"')

        spec = DependencySpecificationDetector(Path("pyproject.toml"), requirements_txt=("req/req.txt",)).detect()
        assert spec == DependencyManagementFormat.REQUIREMENTS_TXT


def test_dependency_specification_not_found_raises_exception(tmp_path: Path) -> None:
    with run_within_dir(tmp_path), pytest.raises(
        DependencySpecificationNotFoundError,
        match=re.escape(
            "No file called 'pyproject.toml' with a [tool.poetry.dependencies], [tool.pdm] or [project] section or"
            " file(s) called 'req/req.txt' found. Exiting."
        ),
    ):
        DependencySpecificationDetector(Path("pyproject.toml"), requirements_txt=("req/req.txt",)).detect()



from __future__ import annotations

from pathlib import Path

from deptry.dependency import Dependency
from deptry.module import ModuleBuilder


def test_simple_import() -> None:
    module = ModuleBuilder("click", {"foo", "bar"}, frozenset()).build()
    assert module.package == "click"
    assert module.standard_library is False
    assert module.local_module is False


def test_top_level() -> None:
    # Test if no error is raised, argument is accepted.
    dependency = Dependency("beautifulsoup4", Path("pyproject.toml"))
    dependency.top_levels = {"bs4"}
    module = ModuleBuilder("bs4", {"foo", "bar"}, frozenset(), [dependency]).build()
    assert module.package is None
    assert module.standard_library is False
    assert module.local_module is False


def test_stdlib() -> None:
    module = ModuleBuilder("sys", {"foo", "bar"}, frozenset({"sys"})).build()
    assert module.package is None
    assert module.standard_library is True
    assert module.local_module is False


def test_local_module() -> None:
    module = ModuleBuilder("click", {"foo", "click"}, frozenset()).build()
    assert module.package is None
    assert module.standard_library is False
    assert module.local_module is True



from __future__ import annotations

import itertools
import re
from dataclasses import dataclass
from typing import TYPE_CHECKING

from deptry.dependency import Dependency
from deptry.dependency_getter.base import DependenciesExtract, DependencyGetter
from deptry.utils import load_pyproject_toml

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence


@dataclass
class PEP621DependencyGetter(DependencyGetter):
    """
    Class to extract dependencies from a pyproject.toml file in which dependencies are specified according to PEP 621. For example:

        [project]
        ...
        dependencies = [
            "httpx",
            "gidgethub[httpx]>4.0.0",
            "django>2.1; os_name != 'nt'",
            "django>2.0; os_name == 'nt'"
        ]

        [project.optional-dependencies]
        test = [
            "pytest < 5.0.0",
            "pytest-cov[all]"
        ]

    Note that both dependencies and optional-dependencies are extracted as regular dependencies. Since PEP-621 does not specify
    a recommended way to extract development dependencies, we do not attempt to extract any from the pyproject.toml file.
    """

    def get(self) -> DependenciesExtract:
        dependencies = [*self._get_dependencies(), *itertools.chain(*self._get_optional_dependencies().values())]
        self._log_dependencies(dependencies)

        return DependenciesExtract(dependencies, [])

    def _get_dependencies(self) -> list[Dependency]:
        pyproject_data = load_pyproject_toml(self.config)
        dependency_strings: list[str] = pyproject_data["project"]["dependencies"]
        return self._extract_pep_508_dependencies(dependency_strings, self.package_module_name_map)

    def _get_optional_dependencies(self) -> dict[str, list[Dependency]]:
        pyproject_data = load_pyproject_toml(self.config)

        return {
            group: self._extract_pep_508_dependencies(dependencies, self.package_module_name_map)
            for group, dependencies in pyproject_data["project"].get("optional-dependencies", {}).items()
        }

    def _extract_pep_508_dependencies(
        self, dependencies: list[str], package_module_name_map: Mapping[str, Sequence[str]]
    ) -> list[Dependency]:
        """
        Given a list of dependency specifications (e.g. "django>2.1; os_name != 'nt'"), convert them to Dependency objects.
        """
        extracted_dependencies = []

        for spec in dependencies:
            # An example of a spec is `"tomli>=1.1.0; python_version < \"3.11\""`
            name = self._find_dependency_name_in(spec)
            if name:
                extracted_dependencies.append(
                    Dependency(
                        name,
                        self.config,
                        conditional=self._is_conditional(spec),
                        optional=self._is_optional(spec),
                        module_names=package_module_name_map.get(name),
                    )
                )

        return extracted_dependencies

    @staticmethod
    def _is_optional(dependency_specification: str) -> bool:
        return bool(re.findall(r"\[([a-zA-Z0-9-]+?)\]", dependency_specification))

    @staticmethod
    def _is_conditional(dependency_specification: str) -> bool:
        return ";" in dependency_specification

    @staticmethod
    def _find_dependency_name_in(spec: str) -> str | None:
        match = re.search("[a-zA-Z0-9-_]+", spec)
        if match:
            return match.group(0)
        return None



from __future__ import annotations

from pathlib import Path

from deptry.utils import load_pyproject_toml


def test_load_pyproject_toml() -> None:
    assert load_pyproject_toml(Path("tests/data/example_project/pyproject.toml")) == {
        "tool": {
            "deptry": {"per_rule_ignores": {"DEP002": ["pkginfo"]}},
            "poetry": {
                "authors": ["test <test@test.com>"],
                "dependencies": {
                    "click": "^8.1.3",
                    "isort": "^5.10.1",
                    "pkginfo": "^1.8.3",
                    "python": ">=3.7,<4.0",
                    "requests": "^2.28.1",
                    "toml": "^0.10.2",
                    "urllib3": "^1.26.12",
                },
                "description": "A test project",
                "dev-dependencies": {"black": "^22.6.0"},
                "name": "test",
                "version": "0.0.1",
            },
        }
    }



from __future__ import annotations

import re
import sys
from pathlib import Path
from unittest import mock

import pytest

from deptry.core import Core
from deptry.dependency import Dependency
from deptry.exceptions import UnsupportedPythonVersionError
from deptry.imports.location import Location
from deptry.module import Module
from deptry.stdlibs import STDLIBS_PYTHON
from deptry.violations import (
    DEP001MissingDependencyViolation,
    DEP002UnusedDependencyViolation,
    DEP003TransitiveDependencyViolation,
    DEP004MisplacedDevDependencyViolation,
)
from tests.utils import create_files, run_within_dir


def test__get_sorted_violations() -> None:
    violations = [
        DEP004MisplacedDevDependencyViolation(Module("foo"), Location(Path("foo.py"), 1, 0)),
        DEP001MissingDependencyViolation(Module("foo"), Location(Path("foo.py"), 2, 0)),
        DEP001MissingDependencyViolation(Module("foo"), Location(Path("foo.py"), 1, 0)),
        DEP001MissingDependencyViolation(Module("foo"), Location(Path("bar.py"), 3, 1)),
        DEP001MissingDependencyViolation(Module("foo"), Location(Path("bar.py"), 2, 1)),
        DEP001MissingDependencyViolation(Module("foo"), Location(Path("bar.py"), 3, 0)),
    ]

    assert Core._get_sorted_violations(violations) == [
        DEP001MissingDependencyViolation(Module("foo"), Location(Path("bar.py"), 2, 1)),
        DEP001MissingDependencyViolation(Module("foo"), Location(Path("bar.py"), 3, 0)),
        DEP001MissingDependencyViolation(Module("foo"), Location(Path("bar.py"), 3, 1)),
        DEP001MissingDependencyViolation(Module("foo"), Location(Path("foo.py"), 1, 0)),
        DEP004MisplacedDevDependencyViolation(Module("foo"), Location(Path("foo.py"), 1, 0)),
        DEP001MissingDependencyViolation(Module("foo"), Location(Path("foo.py"), 2, 0)),
    ]


@pytest.mark.parametrize(
    ("known_first_party", "root_suffix", "expected"),
    [
        (
            (),
            "",
            {"module_with_init", "module_without_init", "local_file"},
        ),
        (
            ("module_with_init", "module_without_init"),
            "",
            {"module_with_init", "module_without_init", "local_file"},
        ),
        (
            ("module_without_init",),
            "module_with_init",
            {"foo", "module_without_init", "subdirectory"},
        ),
    ],
)
def test__get_local_modules(
    tmp_path: Path, known_first_party: tuple[str, ...], root_suffix: str, expected: set[str]
) -> None:
    with run_within_dir(tmp_path):
        create_files([
            Path("module_with_init/__init__.py"),
            Path("module_with_init/foo.py"),
            Path("module_with_init/subdirectory/__init__.py"),
            Path("module_with_init/subdirectory/foo.py"),
            Path("module_without_init/bar.py"),
            Path("local_file.py"),
        ])

        assert (
            Core(
                root=(tmp_path / root_suffix,),
                config=Path("pyproject.toml"),
                no_ansi=False,
                per_rule_ignores={},
                ignore=(),
                exclude=(),
                extend_exclude=(),
                using_default_exclude=True,
                ignore_notebooks=False,
                requirements_txt=(),
                requirements_txt_dev=(),
                known_first_party=known_first_party,
                json_output="",
                package_module_name_map={},
            )._get_local_modules()
            == expected
        )


@pytest.mark.skipif(sys.version_info >= (3, 10), reason="mapping is only used for Python < 3.10")
def test__get_stdlib_packages_without_stdlib_module_names() -> None:
    assert Core._get_stdlib_modules() == STDLIBS_PYTHON[f"{sys.version_info[0]}{sys.version_info[1]}"]


@pytest.mark.skipif(sys.version_info < (3, 10), reason="only Python >= 3.10 has sys.stdlib_module_names")
def test__get_stdlib_packages_with_stdlib_module_names() -> None:
    assert Core._get_stdlib_modules() == sys.stdlib_module_names  # type: ignore[attr-defined, unused-ignore]


@pytest.mark.parametrize(
    "version_info",
    [
        (sys.version_info[0], sys.version_info[1] + 1, 0),
        (sys.version_info[0], sys.version_info[1] + 1, 13),
        (sys.version_info[0] + 1, sys.version_info[1], 0),
        (sys.version_info[0], sys.version_info[1] + 1, 0, "beta", 1),
        (sys.version_info[0], sys.version_info[1] + 1, 0, "candidate", 1),
    ],
)
@pytest.mark.skipif(sys.version_info < (3, 10), reason="only Python >= 3.10 has sys.stdlib_module_names")
def test__get_stdlib_packages_with_stdlib_module_names_future_version(version_info: tuple[int | str, ...]) -> None:
    """Test that future versions of Python not yet tested on the CI will also work."""
    with mock.patch("sys.version_info", (sys.version_info[0], sys.version_info[1] + 1, 0)):
        assert Core._get_stdlib_modules() == sys.stdlib_module_names  # type: ignore[attr-defined, unused-ignore]


@pytest.mark.parametrize(
    "version_info",
    [
        (2, 1, 0),
        (2, 7, 0),
        (2, 7, 15),
        (3, 6, 0),
        (3, 6, 7),
        (3, 6, 7, "candidate", 1),
    ],
)
def test__get_stdlib_packages_unsupported(version_info: tuple[int | str, ...]) -> None:
    """It should raise an error when Python version is unsupported."""
    with mock.patch("sys.version_info", version_info), pytest.raises(
        UnsupportedPythonVersionError,
        match=re.escape(
            f"Python version {version_info[0]}.{version_info[1]} is not supported. Only versions >= 3.8 are supported."
        ),
    ):
        Core._get_stdlib_modules()


def test__exit_with_violations() -> None:
    violations = [
        DEP001MissingDependencyViolation(Module("foo"), Location(Path("foo.py"), 1, 2)),
        DEP002UnusedDependencyViolation(Dependency("foo", Path("pyproject.toml")), Location(Path("pyproject.toml"))),
        DEP003TransitiveDependencyViolation(Module("foo"), Location(Path("foo.py"), 1, 2)),
        DEP004MisplacedDevDependencyViolation(Module("foo"), Location(Path("foo.py"), 1, 2)),
    ]

    with pytest.raises(SystemExit) as e:
        Core._exit(violations)

    assert e.type == SystemExit
    assert e.value.code == 1


def test__exit_without_violations() -> None:
    with pytest.raises(SystemExit) as e:
        Core._exit([])

    assert e.type == SystemExit
    assert e.value.code == 0



from .rust import Location as RustLocation

def get_imports_from_py_files(file_paths: list[str]) -> dict[str, list[RustLocation]]: ...
def get_imports_from_py_file(file_path: str) -> dict[str, list[RustLocation]]: ...

class Location:
    file: str
    line: int
    column: int
    def __init__(self, file: str, line: int, column: int) -> None: ...



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from deptry.deprecate.ignore_flags import generate_deprecation_warning, get_value_for_per_rule_ignores_argument

if TYPE_CHECKING:
    from collections.abc import MutableMapping


def test_generate_deprecation_warning() -> None:
    result = generate_deprecation_warning(flag_name="ignore-missing", issue_code="DEP002", sequence=("hi", "bye"))
    assert (
        result == "Warning: In an upcoming release, support for the `--ignore-missing` command-line option and the"
        " `ignore_missing` configuration parameter will be discontinued. Instead, use `--per-rule-ignores"
        ' DEP002=hi|bye` or add a line `DEP002 = ["hi", "bye"]` to the `[tool.deptry.per_rule_ignores]` section of the'
        " configuration file."
    )


@pytest.mark.parametrize(
    ("ignore_params", "expected_result", "flag_name", "issue_code"),
    [
        (
            {
                "ignore_obsolete": (),
                "ignore_unused": (),
                "ignore_missing": ("hello", "goodbye"),
                "ignore_transitive": (),
                "ignore_misplaced_dev": (),
            },
            {"DEP001": ("hello", "goodbye")},
            "ignore-missing",
            "DEP001",
        ),
        (
            {
                "ignore_obsolete": ("hello", "goodbye"),
                "ignore_unused": (),
                "ignore_missing": (),
                "ignore_transitive": (),
                "ignore_misplaced_dev": (),
            },
            {"DEP002": ("hello", "goodbye")},
            "ignore-obsolete",
            "DEP002",
        ),
        (
            {
                "ignore_obsolete": (),
                "ignore_unused": ("hello", "goodbye"),
                "ignore_missing": (),
                "ignore_transitive": (),
                "ignore_misplaced_dev": (),
            },
            {"DEP002": ("hello", "goodbye")},
            "ignore-unused",
            "DEP002",
        ),
        (
            {
                "ignore_obsolete": (),
                "ignore_unused": (),
                "ignore_missing": (),
                "ignore_transitive": ("hello", "goodbye"),
                "ignore_misplaced_dev": (),
            },
            {"DEP003": ("hello", "goodbye")},
            "ignore-transitive",
            "DEP003",
        ),
        (
            {
                "ignore_obsolete": (),
                "ignore_unused": (),
                "ignore_missing": (),
                "ignore_transitive": (),
                "ignore_misplaced_dev": ("hello", "goodbye"),
            },
            {"DEP004": ("hello", "goodbye")},
            "ignore-misplaced-dev",
            "DEP004",
        ),
    ],
)
def test_ignore_param(
    caplog: pytest.LogCaptureFixture,
    ignore_params: MutableMapping[str, tuple[str, ...]],
    expected_result: MutableMapping[str, tuple[str, ...]],
    flag_name: str,
    issue_code: str,
) -> None:
    result = get_value_for_per_rule_ignores_argument(per_rule_ignores={}, **ignore_params)
    assert result == expected_result
    assert (
        generate_deprecation_warning(flag_name=flag_name, issue_code=issue_code, sequence=("hello", "goodbye"))
        in caplog.text
    )


@pytest.mark.parametrize(
    ("ignore_params", "expected_result", "flag_name", "issue_code"),
    [
        (
            {
                "ignore_obsolete": (),
                "ignore_unused": (),
                "ignore_missing": ("hello", "goodbye"),
                "ignore_transitive": (),
                "ignore_misplaced_dev": (),
            },
            {"DEP001": ("goodbye", "hello", "package")},
            "ignore-missing",
            "DEP001",
        ),
        (
            {
                "ignore_obsolete": ("hello", "goodbye"),
                "ignore_unused": (),
                "ignore_missing": (),
                "ignore_transitive": (),
                "ignore_misplaced_dev": (),
            },
            {"DEP002": ("goodbye", "hello", "package")},
            "ignore-obsolete",
            "DEP002",
        ),
        (
            {
                "ignore_obsolete": (),
                "ignore_unused": ("hello", "goodbye"),
                "ignore_missing": (),
                "ignore_transitive": (),
                "ignore_misplaced_dev": (),
            },
            {"DEP002": ("goodbye", "hello", "package")},
            "ignore-unused",
            "DEP002",
        ),
        (
            {
                "ignore_obsolete": (),
                "ignore_unused": (),
                "ignore_missing": (),
                "ignore_transitive": ("hello", "goodbye"),
                "ignore_misplaced_dev": (),
            },
            {"DEP003": ("goodbye", "hello", "package")},
            "ignore-transitive",
            "DEP003",
        ),
        (
            {
                "ignore_obsolete": (),
                "ignore_unused": (),
                "ignore_missing": (),
                "ignore_transitive": (),
                "ignore_misplaced_dev": ("hello", "goodbye"),
            },
            {"DEP004": ("goodbye", "hello", "package")},
            "ignore-misplaced-dev",
            "DEP004",
        ),
    ],
)
def test_ignore_param_append(
    caplog: pytest.LogCaptureFixture,
    ignore_params: MutableMapping[str, tuple[str, ...]],
    expected_result: MutableMapping[str, tuple[str, ...]],
    flag_name: str,
    issue_code: str,
) -> None:
    result = get_value_for_per_rule_ignores_argument(per_rule_ignores={issue_code: ("package",)}, **ignore_params)
    # We do not care about the order of the sequences that are in the values of the dict.
    assert {k: sorted(v) for k, v in result.items()} == {k: sorted(v) for k, v in expected_result.items()}
    assert (
        generate_deprecation_warning(flag_name=flag_name, issue_code=issue_code, sequence=("hello", "goodbye"))
        in caplog.text
    )



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from pathlib import Path

    from deptry.dependency import Dependency


@dataclass
class DependenciesExtract:
    dependencies: list[Dependency]
    dev_dependencies: list[Dependency]


@dataclass
class DependencyGetter(ABC):
    """Base class for all classes that extract a list of project's dependencies from a file.

    Args:
        config: The path to a configuration file that contains the project's dependencies.
        package_module_name_map: A mapping of package names to their corresponding module names that may not be found
        otherwise from the package's metadata. The keys in the mapping should be package names, and the values should
        be sequences of module names associated with the package.
    """

    config: Path
    package_module_name_map: Mapping[str, Sequence[str]] = field(default_factory=dict)

    @abstractmethod
    def get(self) -> DependenciesExtract:
        """Get extracted dependencies and dev dependencies."""
        raise NotImplementedError()

    @staticmethod
    def _log_dependencies(dependencies: list[Dependency], is_dev: bool = False) -> None:
        logging.debug("The project contains the following %s:", "dev dependencies" if is_dev else "dependencies")

        for dependency in dependencies:
            logging.debug(dependency)

        logging.debug("")



from os import chdir, walk
from pathlib import Path
from typing import List

import numpy as np
import pandas
from numpy.random import sample

from . import foo
from .foo import bar

x = 1
if x > 0:
    import httpx
elif x < 0:
    from baz import Bar
else:
    import foobar

import barfoo as bf
from randomizer import random

try:
    import click
except:
    import not_click


def func():
    import module_in_func


class MyClass:
    def __init__(self):
        import module_in_class



from __future__ import annotations

import logging
import sys
from collections import defaultdict
from importlib.metadata import version
from pathlib import Path
from typing import TYPE_CHECKING

import click

from deptry.config import read_configuration_from_pyproject_toml
from deptry.core import Core
from deptry.deprecate.ignore_flags import get_value_for_per_rule_ignores_argument
from deptry.deprecate.skip_flags import get_value_for_ignore_argument

if TYPE_CHECKING:
    from collections.abc import MutableMapping, Sequence

if sys.platform == "win32":
    from colorama import just_fix_windows_console

    just_fix_windows_console()

DEFAULT_EXCLUDE = ("venv", r"\.venv", r"\.direnv", "tests", r"\.git", r"setup\.py")


class CommaSeparatedTupleParamType(click.ParamType):
    """
    This class is used to uniformly handle configuration parameters that can be either passed as a comma-separated string,
    as a list of strings, or as a tuple of strings. For example, the value for a parameter can be a comma-separated string
    when passed as a command line argument, or as a list of strings when passed through pyproject.toml.
    """

    name = "tuple"

    def convert(
        self,
        value: str | list[str] | tuple[str, ...],
        param: click.Parameter | None,
        ctx: click.Context | None,
    ) -> tuple[str, ...]:
        if isinstance(value, str):
            return tuple(value.split(","))
        if isinstance(value, list):
            return tuple(value)
        return value


COMMA_SEPARATED_TUPLE = CommaSeparatedTupleParamType()


class CommaSeparatedMappingParamType(click.ParamType):
    """
    This class is used to uniformly handle configuration parameters that can be either passed as a comma-separated pair
    string, or as a Mapping of strings to tuples of strings. Items in a pair string are separated by an equal sign,
    where multiple values are separated by a pipe: key1=value1,key2=value2|value3.
    For example, the value for a parameter can be a comma-separated pair string when passed as a command line argument,
    or as a mapping of string to tuples of strings when passed through pyproject.toml.
    """

    name = "mapping"

    def convert(
        self,
        # In the mapping value below, although a str is a Sequence[str] itself,
        # they are treated differently from other sequences of str.
        value: str | MutableMapping[str, Sequence[str] | str],
        param: click.Parameter | None,
        ctx: click.Context | None,
    ) -> dict[str, tuple[str, ...]]:
        converted: dict[str, tuple[str, ...]]
        if isinstance(value, str):
            map_: defaultdict[str, list[str]] = defaultdict(list)
            for item in value.split(","):
                pair = tuple(item.split("=", 1))
                if len(pair) != 2:
                    error_msg = (
                        f"package name and module names pairs should be concatenated with an equal sign (=): {item}"
                    )
                    raise ValueError(error_msg)
                package_name = pair[0]
                module_names = pair[1].split("|")
                map_[package_name].extend(module_names)
            converted = {k: tuple(v) for k, v in map_.items()}
        else:
            converted = {k: (v,) if isinstance(v, str) else tuple(v) for k, v in value.items()}

        return converted


COMMA_SEPARATED_MAPPING = CommaSeparatedMappingParamType()


def configure_logger(_ctx: click.Context, _param: click.Parameter, value: bool) -> None:
    log_level = logging.DEBUG if value else logging.INFO
    logging.basicConfig(level=log_level, handlers=[logging.StreamHandler()], format="%(message)s")


def display_deptry_version(ctx: click.Context, _param: click.Parameter, value: bool) -> None:
    if not value or ctx.resilient_parsing:
        return None

    click.echo(f'deptry {version("deptry")}')
    ctx.exit()


@click.command()
@click.argument("root", type=click.Path(exists=True, path_type=Path), nargs=-1, required=True)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    help=(
        "Boolean flag for verbosity. Using this flag will display more information about files, imports and"
        " dependencies while running."
    ),
    expose_value=False,
    is_eager=True,
    callback=configure_logger,
)
@click.option(
    "--config",
    type=click.Path(path_type=Path),
    is_eager=True,
    callback=read_configuration_from_pyproject_toml,
    help="Path to the pyproject.toml file to read configuration from.",
    default="pyproject.toml",
)
@click.option(
    "--no-ansi",
    is_flag=True,
    help="Disable ANSI characters in terminal output.",
)
@click.option("--skip-obsolete", is_flag=True, hidden=True)
@click.option(
    "--skip-unused",
    is_flag=True,
    help="To be deprecated.",
    hidden=True,
)
@click.option(
    "--skip-missing",
    is_flag=True,
    help="To be deprecated.",
    hidden=True,
)
@click.option(
    "--skip-transitive",
    is_flag=True,
    help="To be deprecated.",
    hidden=True,
)
@click.option(
    "--skip-misplaced-dev",
    is_flag=True,
    help="To be deprecated.",
    hidden=True,
)
@click.option("--ignore-obsolete", "-io", help="To be deprecated.", type=COMMA_SEPARATED_TUPLE, default=(), hidden=True)
@click.option(
    "--ignore-unused",
    "-iu",
    type=COMMA_SEPARATED_TUPLE,
    hidden=True,
    help="To be deprecated.",
    default=(),
)
@click.option(
    "--ignore-missing",
    "-im",
    type=COMMA_SEPARATED_TUPLE,
    hidden=True,
    help="To be deprecated.",
    default=(),
)
@click.option(
    "--ignore-transitive",
    "-it",
    type=COMMA_SEPARATED_TUPLE,
    hidden=True,
    help="To be deprecated.",
    default=(),
)
@click.option(
    "--ignore-misplaced-dev",
    "-id",
    type=COMMA_SEPARATED_TUPLE,
    hidden=True,
    help="To be deprecated.",
    default=(),
)
@click.option(
    "--ignore",
    "-i",
    type=COMMA_SEPARATED_TUPLE,
    help="""A comma-separated list of error codes to ignore. e.g. `deptry --ignore DEP001,DEP002`
    For more information regarding the error codes, see https://fpgmaas.github.io/deptry/issue-codes/""",
    default=(),
)
@click.option(
    "--per-rule-ignores",
    "-pri",
    type=COMMA_SEPARATED_MAPPING,
    help="""A comma-separated mapping of packages or modules to be ignored per error code.
    . e.g. ``deptry . --per-rule-ignores DEP001=matplotlib,DEP002=pandas|numpy``
    For more information regarding the error codes, see https://fpgmaas.github.io/deptry/issue-codes/""",
    default={},
)
@click.option(
    "--exclude",
    "-e",
    multiple=True,
    type=str,
    help=f"""A regular expression for directories or files in which .py files should not be scanned for imports to determine if there are dependency issues.
    Can be used multiple times by specifying the argument multiple times. re.match() is used to match the expressions, which by default checks for a match only at the beginning of a string.
    For example: `deptry . -e ".*/foo/" -e bar"` Note that this overwrites the defaults.
    [default: {", ".join(DEFAULT_EXCLUDE)}
    """,
)
@click.option(
    "--extend-exclude",
    "-ee",
    type=str,
    multiple=True,
    help="""Like --exclude, but adds additional files and directories on top of the excluded ones instead of overwriting the defaults.
    (Useful if you simply want to add to the default) `deptry . -ee ".*/foo/" -ee bar"`""",
    default=(),
    show_default=True,
)
@click.option(
    "--ignore-notebooks",
    "-nb",
    is_flag=True,
    help="Boolean flag to specify if notebooks should be ignored while scanning for imports.",
)
@click.option(
    "--version",
    is_flag=True,
    is_eager=True,
    expose_value=False,
    callback=display_deptry_version,
    help="Display the current version and exit.",
)
@click.option(
    "--requirements-txt",
    "-rt",
    type=COMMA_SEPARATED_TUPLE,
    help=""".txt files to scan for dependencies. If a file called pyproject.toml with a [tool.poetry.dependencies] section is found, this argument is ignored
    and the dependencies are extracted from the pyproject.toml file instead. Can be multiple e.g. `deptry . --requirements-txt req/prod.txt,req/extra.txt`""",
    default=("requirements.txt",),
    show_default=True,
)
@click.option(
    "--requirements-txt-dev",
    "-rtd",
    type=COMMA_SEPARATED_TUPLE,
    help=""".txt files to scan for additional development dependencies. If a file called pyproject.toml with a [tool.poetry.dependencies] section is found, this argument is ignored
    and the dependencies are extracted from the pyproject.toml file instead. Can be multiple e.g. `deptry . --requirements-txt-dev req/dev.txt,req/test.txt`""",
    default=("dev-requirements.txt", "requirements-dev.txt"),
    show_default=True,
)
@click.option(
    "--known-first-party",
    "-kf",
    type=str,
    multiple=True,
    help="Modules to consider as first party ones.",
    default=(),
    show_default=True,
)
@click.option(
    "--json-output",
    "-o",
    type=str,
    help="""If specified, a summary of the dependency issues found will be written to the output location specified. e.g. `deptry . -o deptry.json`""",
    show_default=True,
)
@click.option(
    "--package-module-name-map",
    "-pmnm",
    type=COMMA_SEPARATED_MAPPING,
    help="""Manually defined module names belonging to packages. For example; `deptry . --package-module-name-map package_1=module_a,package_2=module_b|module_c`.""",
    default={},
    show_default=False,
)
def deptry(
    root: tuple[Path, ...],
    config: Path,
    no_ansi: bool,
    ignore_unused: tuple[str, ...],
    ignore_obsolete: tuple[str, ...],
    ignore_missing: tuple[str, ...],
    ignore_transitive: tuple[str, ...],
    ignore_misplaced_dev: tuple[str, ...],
    skip_unused: bool,
    skip_obsolete: bool,
    skip_missing: bool,
    skip_transitive: bool,
    skip_misplaced_dev: bool,
    ignore: tuple[str, ...],
    per_rule_ignores: MutableMapping[str, tuple[str, ...]],
    exclude: tuple[str, ...],
    extend_exclude: tuple[str, ...],
    ignore_notebooks: bool,
    requirements_txt: tuple[str, ...],
    requirements_txt_dev: tuple[str, ...],
    known_first_party: tuple[str, ...],
    json_output: str,
    package_module_name_map: MutableMapping[str, tuple[str, ...]],
) -> None:
    """Find dependency issues in your Python project.

    ROOT is the path to the root directory of the project to be scanned. For instance, to invoke deptry in the current
    directory:

        deptry .

    If your project has multiple source directories, multiple ROOT can be specified.  For instance, to invoke deptry in
    both 'src' and 'worker' directories:

        deptry src worker

    """
    ignore = get_value_for_ignore_argument(
        ignore,
        skip_missing=skip_missing,
        skip_obsolete=skip_obsolete,
        skip_unused=skip_unused,
        skip_transitive=skip_transitive,
        skip_misplaced_dev=skip_misplaced_dev,
    )
    per_rule_ignores = get_value_for_per_rule_ignores_argument(
        per_rule_ignores=per_rule_ignores,
        ignore_missing=ignore_missing,
        ignore_obsolete=ignore_obsolete,
        ignore_unused=ignore_unused,
        ignore_misplaced_dev=ignore_misplaced_dev,
        ignore_transitive=ignore_transitive,
    )
    Core(
        root=root,
        config=config,
        no_ansi=no_ansi,
        exclude=exclude or DEFAULT_EXCLUDE,
        extend_exclude=extend_exclude,
        using_default_exclude=not exclude,
        ignore_notebooks=ignore_notebooks,
        ignore=ignore,
        per_rule_ignores=per_rule_ignores,
        requirements_txt=requirements_txt,
        requirements_txt_dev=requirements_txt_dev,
        known_first_party=known_first_party,
        json_output=json_output,
        package_module_name_map=package_module_name_map,
    ).run()



from __future__ import annotations

import pytest

from tests.utils import PDMVenvFactory, PipVenvFactory, PoetryVenvFactory


@pytest.fixture(scope="session")
def pdm_venv_factory(tmp_path_factory: pytest.TempPathFactory) -> PDMVenvFactory:
    return PDMVenvFactory(tmp_path_factory.getbasetemp() / "venvs")


@pytest.fixture(scope="session")
def poetry_venv_factory(tmp_path_factory: pytest.TempPathFactory) -> PoetryVenvFactory:
    return PoetryVenvFactory(tmp_path_factory.getbasetemp() / "venvs")


@pytest.fixture(scope="session")
def pip_venv_factory(tmp_path_factory: pytest.TempPathFactory) -> PipVenvFactory:
    return PipVenvFactory(tmp_path_factory.getbasetemp() / "venvs")



from __future__ import annotations

import logging

from deptry.violations import (
    DEP001MissingDependencyViolation,
    DEP002UnusedDependencyViolation,
    DEP003TransitiveDependencyViolation,
    DEP004MisplacedDevDependencyViolation,
)


def generate_deprecation_warning(flag_name: str, issue_code: str) -> str:
    return (
        f"Warning: In an upcoming release, support for the `--{flag_name}` command-line option and the"
        f" `{flag_name.replace('-', '_')}` configuration parameter will be discontinued. Instead, use `--ignore"
        f" {issue_code}` or add a line `ignore = ['{issue_code}']` to the `[tool.deptry]` section of the configuration"
        " file."
    )


def get_value_for_ignore_argument(
    ignore: tuple[str, ...],
    skip_obsolete: bool,
    skip_unused: bool,
    skip_missing: bool,
    skip_transitive: bool,
    skip_misplaced_dev: bool,
) -> tuple[str, ...]:
    """
    This function is designed to help with the transition from deprecated command-line flags to the new `--ignore` flag.
    The deprecated flags that are replaced by this new flag are:

        - `--skip-obsolete`
        - `--skip-unused`
        - `--skip-missing`
        - `--skip-transitive`
        - `--skip-misplaced-dev`

    This function accepts the values for the deprecated flags and updates the `ignore` parameter accordingly.

    Raise a warning if one of the to-be-deprecated flags is used.
    """
    user_values = {
        "skip-missing": skip_missing,
        "skip-unused": skip_unused,
        "skip-obsolete": skip_obsolete,
        "skip-transitive": skip_transitive,
        "skip-misplaced-dev": skip_misplaced_dev,
    }

    issue_codes = {
        "skip-missing": DEP001MissingDependencyViolation.error_code,
        "skip-unused": DEP002UnusedDependencyViolation.error_code,
        "skip-obsolete": DEP002UnusedDependencyViolation.error_code,
        "skip-transitive": DEP003TransitiveDependencyViolation.error_code,
        "skip-misplaced-dev": DEP004MisplacedDevDependencyViolation.error_code,
    }

    for flag, should_skip in user_values.items():
        if should_skip:
            code = issue_codes[flag]
            logging.warning(generate_deprecation_warning(flag, code))
            if code not in ignore:
                ignore += (code,)

    return ignore



#!/usr/bin/env python3

# This script is inspired by isort: https://github.com/PyCQA/isort/blob/4ccbd1eddf564d2c9e79c59d59c1fc06a7e35f94/scripts/mkstdlibs.py.

from __future__ import annotations

import ast
import urllib.request
from html.parser import HTMLParser
from pathlib import Path

OUTPUT_PATH = Path("deptry/stdlibs.py")
STDLIB_MODULES_URL = "https://docs.python.org/{}.{}/py-modindex.html"

# Starting from Python 3.10, https://docs.python.org/3/library/sys.html#sys.stdlib_module_names is available.
PYTHON_VERSIONS = ((3, 8), (3, 9))

# Modules that are in stdlib, but undocumented.
EXTRA_STDLIBS_MODULES = ("_ast", "ntpath", "posixpath", "sre", "sre_constants", "sre_compile", "sre_parse")

DOCSTRING_GENERATED_FILES = """
DO NOT EDIT THIS FILE MANUALLY.
It is generated from `scripts/generate_stdlibs.py` script and contains the stdlib modules for Python versions that does
not support https://docs.python.org/3/library/sys.html#sys.stdlib_module_names (< 3.10).
The file can be generated again using `python scripts/generate_stdlibs.py`.
"""


class PythonStdlibHTMLParser(HTMLParser):
    def __init__(self) -> None:
        super().__init__()
        self._is_in_code_tag = False
        self.modules: list[str] = []

    def handle_starttag(self, tag: str, attrs: list[tuple[str, str | None]]) -> None:
        if tag == "code":
            self._is_in_code_tag = True

    def handle_endtag(self, tag: str) -> None:
        if tag == "code":
            self._is_in_code_tag = False

    def handle_data(self, data: str) -> None:
        if self._is_in_code_tag:
            self.modules.append(data)


def get_stdlib_modules_for_python_version(python_version: tuple[int, int]) -> list[str]:
    with urllib.request.urlopen(  # noqa: S310
        STDLIB_MODULES_URL.format(python_version[0], python_version[1])
    ) as response:
        html_content = response.read().decode()

    parser = PythonStdlibHTMLParser()
    parser.feed(html_content)

    modules = {module.split(".")[0] for module in parser.modules}.union(EXTRA_STDLIBS_MODULES)
    modules.remove("__main__")

    return sorted(modules)


def get_stdlib_modules() -> dict[str, list[str]]:
    return {
        f"{python_version[0]}{python_version[1]}": get_stdlib_modules_for_python_version(python_version)
        for python_version in PYTHON_VERSIONS
    }


def write_stdlibs_file(stdlib_python: dict[str, list[str]]) -> None:
    node = ast.Module(
        body=[
            ast.Expr(ast.Constant(DOCSTRING_GENERATED_FILES)),
            ast.Assign(
                targets=[ast.Name("STDLIBS_PYTHON")],
                value=ast.Dict(
                    keys=[ast.Str(python_version) for python_version in stdlib_python],
                    values=[
                        ast.Call(
                            func=ast.Name(id="frozenset"),
                            args=[ast.Set(elts=[ast.Constant(module) for module in python_stdlib_modules])],
                            keywords=[],
                        )
                        for python_stdlib_modules in stdlib_python.values()
                    ],
                ),
                lineno=None,
            ),
        ],
        type_ignores=[],
    )

    with OUTPUT_PATH.open("w+") as stdlib_file:
        stdlib_file.write(ast.unparse(node))  # type: ignore[attr-defined, unused-ignore]


if __name__ == "__main__":
    write_stdlibs_file(get_stdlib_modules())



from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from pathlib import Path


class ProjectBuilder(Protocol):
    def __call__(self, root_directory: Path, project: str, setup_commands: list[str], cwd: str | None) -> Path: ...


class ToolSpecificProjectBuilder(Protocol):
    def __call__(self, root_directory: Path, project: str, cwd: str | None = None) -> Path: ...



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING

from deptry.imports.location import Location
from deptry.violations.base import ViolationsFinder
from deptry.violations.dep002_unused.violation import DEP002UnusedDependencyViolation

if TYPE_CHECKING:
    from deptry.dependency import Dependency
    from deptry.violations import Violation


@dataclass
class DEP002UnusedDependenciesFinder(ViolationsFinder):
    """
    Finds unused dependencies by comparing a list of imported modules to a list of project dependencies.

    A dependency is considered unused if none of the following conditions hold:
    - A module with the exact name of the dependency is imported.
    - Any of the top-level modules of the dependency are imported.

    For example, 'matplotlib' has top-levels ['matplotlib', 'mpl_toolkits', 'pylab']. `mpl_toolkits` does not have
    any associated metadata, but if this is imported the associated dependency `matplotlib` is not unused,
    even if `matplotlib` itself is not imported anywhere.
    """

    def find(self) -> list[Violation]:
        logging.debug("\nScanning for unused dependencies...")
        unused_dependencies: list[Violation] = []

        for dependency in self.dependencies:
            logging.debug("Scanning module %s...", dependency.name)

            if self._is_unused(dependency):
                unused_dependencies.append(
                    DEP002UnusedDependencyViolation(dependency, Location(dependency.definition_file))
                )

        return unused_dependencies

    def _is_unused(self, dependency: Dependency) -> bool:
        if self._dependency_found_in_imported_modules(dependency) or self._any_of_the_top_levels_imported(dependency):
            return False

        if dependency.name in self.ignored_modules:
            logging.debug("Dependency '%s' found to be unused, but ignoring.", dependency.name)
            return False

        logging.debug("Dependency '%s' does not seem to be used.", dependency.name)
        return True

    def _dependency_found_in_imported_modules(self, dependency: Dependency) -> bool:
        return any(
            module_with_locations.module.package == dependency.name
            for module_with_locations in self.imported_modules_with_locations
        )

    def _any_of_the_top_levels_imported(self, dependency: Dependency) -> bool:
        if not dependency.top_levels:
            return False

        return any(
            any(
                module_with_locations.module.name == top_level
                for module_with_locations in self.imported_modules_with_locations
            )
            for top_level in dependency.top_levels
        )



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import pytest

from deptry.deprecate.skip_flags import generate_deprecation_warning, get_value_for_ignore_argument


def test_generate_deprecation_warning() -> None:
    result = generate_deprecation_warning(flag_name="skip-unused", issue_code="DEP002")
    assert (
        result
        == "Warning: In an upcoming release, support for the `--skip-unused` command-line option and the `skip_unused`"
        " configuration parameter will be discontinued. Instead, use `--ignore DEP002` or add a line `ignore ="
        " ['DEP002']` to the `[tool.deptry]` section of the configuration file."
    )


def test_skip_obsolete(caplog: pytest.LogCaptureFixture) -> None:
    result = get_value_for_ignore_argument(
        ignore=(),
        skip_obsolete=True,
        skip_unused=False,
        skip_missing=False,
        skip_transitive=False,
        skip_misplaced_dev=False,
    )
    assert result == ("DEP002",)
    assert generate_deprecation_warning(flag_name="skip-obsolete", issue_code="DEP002") in caplog.text


def test_skip_unused(caplog: pytest.LogCaptureFixture) -> None:
    result = get_value_for_ignore_argument(
        ignore=(),
        skip_obsolete=True,
        skip_unused=True,
        skip_missing=False,
        skip_transitive=False,
        skip_misplaced_dev=False,
    )
    assert result == ("DEP002",)
    assert generate_deprecation_warning(flag_name="skip-unused", issue_code="DEP002") in caplog.text


def test_skip_transitive_and_missing(caplog: pytest.LogCaptureFixture) -> None:
    result = get_value_for_ignore_argument(
        ignore=(),
        skip_obsolete=False,
        skip_unused=False,
        skip_missing=True,
        skip_transitive=True,
        skip_misplaced_dev=False,
    )
    assert result == ("DEP001", "DEP003")
    assert generate_deprecation_warning(flag_name="skip-missing", issue_code="DEP001") in caplog.text
    assert generate_deprecation_warning(flag_name="skip-transitive", issue_code="DEP003") in caplog.text



from __future__ import annotations

import logging
import operator
import sys
from dataclasses import dataclass
from typing import TYPE_CHECKING

from deptry.dependency_getter.pdm import PDMDependencyGetter
from deptry.dependency_getter.pep_621 import PEP621DependencyGetter
from deptry.dependency_getter.poetry import PoetryDependencyGetter
from deptry.dependency_getter.requirements_txt import RequirementsTxtDependencyGetter
from deptry.dependency_specification_detector import DependencyManagementFormat, DependencySpecificationDetector
from deptry.exceptions import IncorrectDependencyFormatError, UnsupportedPythonVersionError
from deptry.imports.extract import get_imported_modules_from_list_of_files
from deptry.module import ModuleBuilder, ModuleLocations
from deptry.python_file_finder import PythonFileFinder
from deptry.reporters import JSONReporter, TextReporter
from deptry.stdlibs import STDLIBS_PYTHON
from deptry.violations import (
    DEP001MissingDependenciesFinder,
    DEP001MissingDependencyViolation,
    DEP002UnusedDependenciesFinder,
    DEP002UnusedDependencyViolation,
    DEP003TransitiveDependenciesFinder,
    DEP003TransitiveDependencyViolation,
    DEP004MisplacedDevDependenciesFinder,
    DEP004MisplacedDevDependencyViolation,
)

if TYPE_CHECKING:
    from collections.abc import Mapping
    from pathlib import Path

    from deptry.dependency import Dependency
    from deptry.dependency_getter.base import DependenciesExtract
    from deptry.violations import Violation


@dataclass
class Core:
    root: tuple[Path, ...]
    config: Path
    no_ansi: bool
    per_rule_ignores: Mapping[str, tuple[str, ...]]
    ignore: tuple[str, ...]
    exclude: tuple[str, ...]
    extend_exclude: tuple[str, ...]
    using_default_exclude: bool
    ignore_notebooks: bool
    requirements_txt: tuple[str, ...]
    requirements_txt_dev: tuple[str, ...]
    known_first_party: tuple[str, ...]
    json_output: str
    package_module_name_map: Mapping[str, tuple[str, ...]]

    def run(self) -> None:
        self._log_config()

        dependency_management_format = DependencySpecificationDetector(
            self.config, requirements_txt=self.requirements_txt
        ).detect()
        dependencies_extract = self._get_dependencies(dependency_management_format)

        all_python_files = PythonFileFinder(
            self.exclude, self.extend_exclude, self.using_default_exclude, self.ignore_notebooks
        ).get_all_python_files_in(self.root)

        local_modules = self._get_local_modules()
        stdlib_modules = self._get_stdlib_modules()

        imported_modules_with_locations = [
            ModuleLocations(
                ModuleBuilder(
                    module,
                    local_modules,
                    stdlib_modules,
                    dependencies_extract.dependencies,
                    dependencies_extract.dev_dependencies,
                ).build(),
                locations,
            )
            for module, locations in get_imported_modules_from_list_of_files(all_python_files).items()
        ]
        imported_modules_with_locations = [
            module_with_locations
            for module_with_locations in imported_modules_with_locations
            if not module_with_locations.module.standard_library
        ]

        violations = self._find_violations(imported_modules_with_locations, dependencies_extract.dependencies)
        TextReporter(violations, use_ansi=not self.no_ansi).report()

        if self.json_output:
            JSONReporter(violations, self.json_output).report()

        self._exit(violations)

    def _find_violations(
        self, imported_modules_with_locations: list[ModuleLocations], dependencies: list[Dependency]
    ) -> list[Violation]:
        violations = []

        if DEP001MissingDependencyViolation.error_code not in self.ignore:
            violations.extend(
                DEP001MissingDependenciesFinder(
                    imported_modules_with_locations, dependencies, self.per_rule_ignores.get("DEP001", ())
                ).find()
            )

        if DEP002UnusedDependencyViolation.error_code not in self.ignore:
            violations.extend(
                DEP002UnusedDependenciesFinder(
                    imported_modules_with_locations, dependencies, self.per_rule_ignores.get("DEP002", ())
                ).find()
            )

        if DEP003TransitiveDependencyViolation.error_code not in self.ignore:
            violations.extend(
                DEP003TransitiveDependenciesFinder(
                    imported_modules_with_locations, dependencies, self.per_rule_ignores.get("DEP003", ())
                ).find()
            )

        if DEP004MisplacedDevDependencyViolation.error_code not in self.ignore:
            violations.extend(
                DEP004MisplacedDevDependenciesFinder(
                    imported_modules_with_locations, dependencies, self.per_rule_ignores.get("DEP004", ())
                ).find()
            )

        return self._get_sorted_violations(violations)

    @staticmethod
    def _get_sorted_violations(violations: list[Violation]) -> list[Violation]:
        return sorted(
            violations, key=operator.attrgetter("location.file", "location.line", "location.column", "error_code")
        )

    def _get_dependencies(self, dependency_management_format: DependencyManagementFormat) -> DependenciesExtract:
        if dependency_management_format is DependencyManagementFormat.POETRY:
            return PoetryDependencyGetter(self.config, self.package_module_name_map).get()
        if dependency_management_format is DependencyManagementFormat.PDM:
            return PDMDependencyGetter(self.config, self.package_module_name_map).get()
        if dependency_management_format is DependencyManagementFormat.PEP_621:
            return PEP621DependencyGetter(self.config, self.package_module_name_map).get()
        if dependency_management_format is DependencyManagementFormat.REQUIREMENTS_TXT:
            return RequirementsTxtDependencyGetter(
                self.config, self.package_module_name_map, self.requirements_txt, self.requirements_txt_dev
            ).get()
        raise IncorrectDependencyFormatError

    def _get_local_modules(self) -> set[str]:
        """
        Get all local Python modules from the source directories and `known_first_party` list.
        A module is considered a local Python module if it matches at least one of those conditions:
        - it is a directory that contains at least one Python file
        - it is a Python file that is not named `__init__.py` (since it is a special case)
        - it is set in the `known_first_party` list
        """
        guessed_local_modules = {
            path.stem for source in self.root for path in source.iterdir() if self._is_local_module(path)
        }

        return guessed_local_modules | set(self.known_first_party)

    @staticmethod
    def _is_local_module(path: Path) -> bool:
        """Guess if a module is a local Python module."""
        return bool(
            (path.is_file() and path.name != "__init__.py" and path.suffix == ".py")
            or (path.is_dir() and list(path.glob("*.py")))
        )

    @staticmethod
    def _get_stdlib_modules() -> frozenset[str]:
        if sys.version_info[:2] >= (3, 10):
            return sys.stdlib_module_names

        try:  # type: ignore[unreachable, unused-ignore]
            return STDLIBS_PYTHON[f"{sys.version_info[0]}{sys.version_info[1]}"]
        except KeyError as e:
            raise UnsupportedPythonVersionError((sys.version_info[0], sys.version_info[1])) from e

    def _log_config(self) -> None:
        logging.debug("Running with the following configuration:")
        for key, value in vars(self).items():
            logging.debug("%s: %s", key, value)
        logging.debug("")

    @staticmethod
    def _exit(violations: list[Violation]) -> None:
        sys.exit(bool(violations))



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from deptry.violations import (
    DEP001MissingDependencyViolation,
    DEP002UnusedDependencyViolation,
    DEP003TransitiveDependencyViolation,
    DEP004MisplacedDevDependencyViolation,
)

if TYPE_CHECKING:
    from collections.abc import MutableMapping


def generate_deprecation_warning(flag_name: str, issue_code: str, sequence: tuple[str, ...]) -> str:
    sequence_as_list_string = "[" + ", ".join(f'"{x}"' for x in sequence) + "]"
    return (
        f"Warning: In an upcoming release, support for the `--{flag_name}` command-line option and the"
        f" `{flag_name.replace('-', '_')}` configuration parameter will be discontinued. Instead, use"
        f" `--per-rule-ignores {issue_code}={'|'.join(sequence)}` or add a line `{issue_code} ="
        f" {sequence_as_list_string}` to the `[tool.deptry.per_rule_ignores]` section of the configuration file."
    )


def get_value_for_per_rule_ignores_argument(
    per_rule_ignores: MutableMapping[str, tuple[str, ...]],
    ignore_obsolete: tuple[str, ...],
    ignore_unused: tuple[str, ...],
    ignore_missing: tuple[str, ...],
    ignore_transitive: tuple[str, ...],
    ignore_misplaced_dev: tuple[str, ...],
) -> MutableMapping[str, tuple[str, ...]]:
    """
    This function is designed to help with the transition from deprecated command-line flags to the new `--per-rule-ignores` flag.
    The deprecated flags that are replaced by this new flag are:

        - `--ignore-obsolete`
        - `--ignore-unused`
        - `--ignore-missing`
        - `--ignore-transitive`
        - `--ignore-misplaced-dev`

    This function accepts the values for the deprecated flags and updates the `per_rule_ignores` mapping accordingly.

    Raise a warning if one of the to-be-deprecated flags is used.
    """
    user_values = {
        "ignore-missing": ignore_missing,
        "ignore-unused": ignore_unused,
        "ignore-obsolete": ignore_obsolete,
        "ignore-transitive": ignore_transitive,
        "ignore-misplaced-dev": ignore_misplaced_dev,
    }

    issue_codes = {
        "ignore-missing": DEP001MissingDependencyViolation.error_code,
        "ignore-unused": DEP002UnusedDependencyViolation.error_code,
        "ignore-obsolete": DEP002UnusedDependencyViolation.error_code,
        "ignore-transitive": DEP003TransitiveDependencyViolation.error_code,
        "ignore-misplaced-dev": DEP004MisplacedDevDependencyViolation.error_code,
    }

    for flag, modules_or_dependencies_to_be_ignored in user_values.items():
        if modules_or_dependencies_to_be_ignored:
            code = issue_codes[flag]
            logging.warning(generate_deprecation_warning(flag, code, modules_or_dependencies_to_be_ignored))
            if code not in per_rule_ignores:
                per_rule_ignores[code] = modules_or_dependencies_to_be_ignored
            else:
                per_rule_ignores[code] = tuple(set(per_rule_ignores[code]).union(modules_or_dependencies_to_be_ignored))

    return per_rule_ignores



from __future__ import annotations

from deptry.violations.base import Violation
from deptry.violations.dep001_missing.finder import DEP001MissingDependenciesFinder
from deptry.violations.dep001_missing.violation import DEP001MissingDependencyViolation
from deptry.violations.dep002_unused.finder import DEP002UnusedDependenciesFinder
from deptry.violations.dep002_unused.violation import DEP002UnusedDependencyViolation
from deptry.violations.dep003_transitive.finder import DEP003TransitiveDependenciesFinder
from deptry.violations.dep003_transitive.violation import DEP003TransitiveDependencyViolation
from deptry.violations.dep004_misplaced_dev.finder import DEP004MisplacedDevDependenciesFinder
from deptry.violations.dep004_misplaced_dev.violation import DEP004MisplacedDevDependencyViolation

__all__ = (
    "DEP001MissingDependencyViolation",
    "DEP002UnusedDependencyViolation",
    "DEP003TransitiveDependencyViolation",
    "DEP004MisplacedDevDependencyViolation",
    "DEP001MissingDependenciesFinder",
    "DEP002UnusedDependenciesFinder",
    "DEP003TransitiveDependenciesFinder",
    "DEP004MisplacedDevDependenciesFinder",
    "Violation",
    "ViolationFinder",
)



from __future__ import annotations

import uuid
from pathlib import Path
from typing import TYPE_CHECKING

from click.testing import CliRunner

from deptry.cli import deptry
from tests.utils import get_issues_report, stylize

if TYPE_CHECKING:
    from tests.utils import PoetryVenvFactory


def test_cli_returns_error(poetry_venv_factory: PoetryVenvFactory) -> None:
    with poetry_venv_factory("example_project") as virtual_env:
        issue_report = f"{uuid.uuid4()}.json"
        result = virtual_env.run(f"deptry . -o {issue_report}")

        assert result.returncode == 1
        assert get_issues_report(Path(issue_report)) == [
            {
                "error": {
                    "code": "DEP002",
                    "message": "'isort' defined as a dependency but not used in the codebase",
                },
                "module": "isort",
                "location": {
                    "file": str(Path("pyproject.toml")),
                    "line": None,
                    "column": None,
                },
            },
            {
                "error": {
                    "code": "DEP002",
                    "message": "'requests' defined as a dependency but not used in the codebase",
                },
                "module": "requests",
                "location": {
                    "file": str(Path("pyproject.toml")),
                    "line": None,
                    "column": None,
                },
            },
            {
                "error": {
                    "code": "DEP004",
                    "message": "'black' imported but declared as a dev dependency",
                },
                "module": "black",
                "location": {
                    "file": str(Path("src/main.py")),
                    "line": 4,
                    "column": 8,
                },
            },
            {
                "error": {
                    "code": "DEP001",
                    "message": "'white' imported but missing from the dependency definitions",
                },
                "module": "white",
                "location": {
                    "file": str(Path("src/main.py")),
                    "line": 6,
                    "column": 8,
                },
            },
        ]


def test_cli_ignore_notebooks(poetry_venv_factory: PoetryVenvFactory) -> None:
    with poetry_venv_factory("example_project") as virtual_env:
        issue_report = f"{uuid.uuid4()}.json"
        result = virtual_env.run(f"deptry . --ignore-notebooks -o {issue_report}")

        assert result.returncode == 1
        assert get_issues_report(Path(issue_report)) == [
            {
                "error": {
                    "code": "DEP002",
                    "message": "'toml' defined as a dependency but not used in the codebase",
                },
                "module": "toml",
                "location": {
                    "file": str(Path("pyproject.toml")),
                    "line": None,
                    "column": None,
                },
            },
            {
                "error": {
                    "code": "DEP002",
                    "message": "'isort' defined as a dependency but not used in the codebase",
                },
                "module": "isort",
                "location": {
                    "file": str(Path("pyproject.toml")),
                    "line": None,
                    "column": None,
                },
            },
            {
                "error": {
                    "code": "DEP002",
                    "message": "'requests' defined as a dependency but not used in the codebase",
                },
                "module": "requests",
                "location": {
                    "file": str(Path("pyproject.toml")),
                    "line": None,
                    "column": None,
                },
            },
            {
                "error": {
                    "code": "DEP004",
                    "message": "'black' imported but declared as a dev dependency",
                },
                "module": "black",
                "location": {
                    "file": str(Path("src/main.py")),
                    "line": 4,
                    "column": 8,
                },
            },
            {
                "error": {
                    "code": "DEP001",
                    "message": "'white' imported but missing from the dependency definitions",
                },
                "module": "white",
                "location": {
                    "file": str(Path("src/main.py")),
                    "line": 6,
                    "column": 8,
                },
            },
        ]


def test_cli_ignore_flags(poetry_venv_factory: PoetryVenvFactory) -> None:
    with poetry_venv_factory("example_project") as virtual_env:
        result = virtual_env.run("deptry . --per-rule-ignores DEP002=isort|pkginfo|requests -im white -id black")

        assert result.returncode == 0


def test_cli_ignore_flag(poetry_venv_factory: PoetryVenvFactory) -> None:
    with poetry_venv_factory("example_project") as virtual_env:
        result = virtual_env.run("deptry . --ignore DEP001,DEP002,DEP003,DEP004")

        assert result.returncode == 0


def test_cli_exclude(poetry_venv_factory: PoetryVenvFactory) -> None:
    with poetry_venv_factory("example_project") as virtual_env:
        issue_report = f"{uuid.uuid4()}.json"
        result = virtual_env.run(f"deptry . --exclude src/notebook.ipynb -o {issue_report}")

        assert result.returncode == 1
        assert get_issues_report(Path(issue_report)) == [
            {
                "error": {
                    "code": "DEP002",
                    "message": "'toml' defined as a dependency but not used in the codebase",
                },
                "module": "toml",
                "location": {
                    "file": str(Path("pyproject.toml")),
                    "line": None,
                    "column": None,
                },
            },
            {
                "error": {
                    "code": "DEP002",
                    "message": "'isort' defined as a dependency but not used in the codebase",
                },
                "module": "isort",
                "location": {
                    "file": str(Path("pyproject.toml")),
                    "line": None,
                    "column": None,
                },
            },
            {
                "error": {
                    "code": "DEP002",
                    "message": "'requests' defined as a dependency but not used in the codebase",
                },
                "module": "requests",
                "location": {
                    "file": str(Path("pyproject.toml")),
                    "line": None,
                    "column": None,
                },
            },
            {
                "error": {
                    "code": "DEP004",
                    "message": "'black' imported but declared as a dev dependency",
                },
                "module": "black",
                "location": {
                    "file": str(Path("src/main.py")),
                    "line": 4,
                    "column": 8,
                },
            },
            {
                "error": {
                    "code": "DEP001",
                    "message": "'white' imported but missing from the dependency definitions",
                },
                "module": "white",
                "location": {
                    "file": str(Path("src/main.py")),
                    "line": 6,
                    "column": 8,
                },
            },
        ]


def test_cli_extend_exclude(poetry_venv_factory: PoetryVenvFactory) -> None:
    with poetry_venv_factory("example_project") as virtual_env:
        issue_report = f"{uuid.uuid4()}.json"
        result = virtual_env.run(f"deptry . -ee src/notebook.ipynb -o {issue_report}")

        assert result.returncode == 1
        assert get_issues_report(Path(issue_report)) == [
            {
                "error": {
                    "code": "DEP002",
                    "message": "'toml' defined as a dependency but not used in the codebase",
                },
                "module": "toml",
                "location": {
                    "file": str(Path("pyproject.toml")),
                    "line": None,
                    "column": None,
                },
            },
            {
                "error": {
                    "code": "DEP002",
                    "message": "'isort' defined as a dependency but not used in the codebase",
                },
                "module": "isort",
                "location": {
                    "file": str(Path("pyproject.toml")),
                    "line": None,
                    "column": None,
                },
            },
            {
                "error": {
                    "code": "DEP002",
                    "message": "'requests' defined as a dependency but not used in the codebase",
                },
                "module": "requests",
                "location": {
                    "file": str(Path("pyproject.toml")),
                    "line": None,
                    "column": None,
                },
            },
            {
                "error": {
                    "code": "DEP004",
                    "message": "'black' imported but declared as a dev dependency",
                },
                "module": "black",
                "location": {
                    "file": str(Path("src/main.py")),
                    "line": 4,
                    "column": 8,
                },
            },
            {
                "error": {
                    "code": "DEP001",
                    "message": "'white' imported but missing from the dependency definitions",
                },
                "module": "white",
                "location": {
                    "file": str(Path("src/main.py")),
                    "line": 6,
                    "column": 8,
                },
            },
        ]


def test_cli_known_first_party(poetry_venv_factory: PoetryVenvFactory) -> None:
    with poetry_venv_factory("example_project") as virtual_env:
        issue_report = f"{uuid.uuid4()}.json"
        result = virtual_env.run(f"deptry . --known-first-party white -o {issue_report}")

        assert result.returncode == 1
        assert get_issues_report(Path(issue_report)) == [
            {
                "error": {
                    "code": "DEP002",
                    "message": "'isort' defined as a dependency but not used in the codebase",
                },
                "module": "isort",
                "location": {
                    "file": str(Path("pyproject.toml")),
                    "line": None,
                    "column": None,
                },
            },
            {
                "error": {
                    "code": "DEP002",
                    "message": "'requests' defined as a dependency but not used in the codebase",
                },
                "module": "requests",
                "location": {
                    "file": str(Path("pyproject.toml")),
                    "line": None,
                    "column": None,
                },
            },
            {
                "error": {
                    "code": "DEP004",
                    "message": "'black' imported but declared as a dev dependency",
                },
                "module": "black",
                "location": {
                    "file": str(Path("src/main.py")),
                    "line": 4,
                    "column": 8,
                },
            },
        ]


def test_cli_not_verbose(poetry_venv_factory: PoetryVenvFactory) -> None:
    with poetry_venv_factory("example_project") as virtual_env:
        issue_report = f"{uuid.uuid4()}.json"
        result = virtual_env.run(f"deptry . -o {issue_report}")

        assert result.returncode == 1
        assert "The project contains the following dependencies:" not in result.stderr
        assert get_issues_report(Path(issue_report)) == [
            {
                "error": {
                    "code": "DEP002",
                    "message": "'isort' defined as a dependency but not used in the codebase",
                },
                "module": "isort",
                "location": {
                    "file": str(Path("pyproject.toml")),
                    "line": None,
                    "column": None,
                },
            },
            {
                "error": {
                    "code": "DEP002",
                    "message": "'requests' defined as a dependency but not used in the codebase",
                },
                "module": "requests",
                "location": {
                    "file": str(Path("pyproject.toml")),
                    "line": None,
                    "column": None,
                },
            },
            {
                "error": {
                    "code": "DEP004",
                    "message": "'black' imported but declared as a dev dependency",
                },
                "module": "black",
                "location": {
                    "file": str(Path("src/main.py")),
                    "line": 4,
                    "column": 8,
                },
            },
            {
                "error": {
                    "code": "DEP001",
                    "message": "'white' imported but missing from the dependency definitions",
                },
                "module": "white",
                "location": {
                    "file": str(Path("src/main.py")),
                    "line": 6,
                    "column": 8,
                },
            },
        ]


def test_cli_verbose(poetry_venv_factory: PoetryVenvFactory) -> None:
    with poetry_venv_factory("example_project") as virtual_env:
        issue_report = f"{uuid.uuid4()}.json"
        result = virtual_env.run(f"deptry . --verbose -o {issue_report}")

        assert result.returncode == 1
        assert "The project contains the following dependencies:" in result.stderr
        assert get_issues_report(Path(issue_report)) == [
            {
                "error": {
                    "code": "DEP002",
                    "message": "'isort' defined as a dependency but not used in the codebase",
                },
                "module": "isort",
                "location": {
                    "file": str(Path("pyproject.toml")),
                    "line": None,
                    "column": None,
                },
            },
            {
                "error": {
                    "code": "DEP002",
                    "message": "'requests' defined as a dependency but not used in the codebase",
                },
                "module": "requests",
                "location": {
                    "file": str(Path("pyproject.toml")),
                    "line": None,
                    "column": None,
                },
            },
            {
                "error": {
                    "code": "DEP004",
                    "message": "'black' imported but declared as a dev dependency",
                },
                "module": "black",
                "location": {
                    "file": str(Path("src/main.py")),
                    "line": 4,
                    "column": 8,
                },
            },
            {
                "error": {
                    "code": "DEP001",
                    "message": "'white' imported but missing from the dependency definitions",
                },
                "module": "white",
                "location": {
                    "file": str(Path("src/main.py")),
                    "line": 6,
                    "column": 8,
                },
            },
        ]


def test_cli_with_no_ansi(poetry_venv_factory: PoetryVenvFactory) -> None:
    with poetry_venv_factory("example_project") as virtual_env:
        result = virtual_env.run("deptry . --no-ansi")

        expected_output = [
            "Scanning 2 files...",
            "",
            f"{Path('pyproject.toml')}: DEP002 'isort' defined as a dependency but not used in the codebase",
            f"{Path('pyproject.toml')}: DEP002 'requests' defined as a dependency but not used in the codebase",
            f"{Path('src/main.py')}:4:8: DEP004 'black' imported but declared as a dev dependency",
            f"{Path('src/main.py')}:6:8: DEP001 'white' imported but missing from the dependency definitions",
            "Found 4 dependency issues.",
            "",
            "For more information, see the documentation: https://fpgmaas.github.io/deptry/",
            "",
        ]

        assert result.returncode == 1
        assert result.stderr == "\n".join(expected_output)


def test_cli_with_not_json_output(poetry_venv_factory: PoetryVenvFactory) -> None:
    with poetry_venv_factory("example_project") as virtual_env:
        json_files_count = len(list(Path().glob("*.json")))

        result = virtual_env.run("deptry .")

        expected_output = [
            "Scanning 2 files...",
            "",
            stylize(
                "{BOLD}{file}{RESET}{CYAN}:{RESET} {BOLD}{RED}DEP002{RESET} 'isort' defined as a dependency but not"
                " used in the codebase",
                file=Path("pyproject.toml"),
            ),
            stylize(
                "{BOLD}{file}{RESET}{CYAN}:{RESET} {BOLD}{RED}DEP002{RESET} 'requests' defined as a dependency but"
                " not used in the codebase",
                file=Path("pyproject.toml"),
            ),
            stylize(
                "{BOLD}{file}{RESET}{CYAN}:{RESET}4{CYAN}:{RESET}8{CYAN}:{RESET} {BOLD}{RED}DEP004{RESET} 'black'"
                " imported but declared as a dev dependency",
                file=Path("src/main.py"),
            ),
            stylize(
                "{BOLD}{file}{RESET}{CYAN}:{RESET}6{CYAN}:{RESET}8{CYAN}:{RESET} {BOLD}{RED}DEP001{RESET} 'white'"
                " imported but missing from the dependency definitions",
                file=Path("src/main.py"),
            ),
            stylize("{BOLD}{RED}Found 4 dependency issues.{RESET}"),
            "",
            "For more information, see the documentation: https://fpgmaas.github.io/deptry/",
            "",
        ]

        assert result.returncode == 1
        # Assert that we have the same number of JSON files as before running the command.
        assert len(list(Path().glob("*.json"))) == json_files_count
        assert result.stderr == "\n".join(expected_output)


def test_cli_with_json_output(poetry_venv_factory: PoetryVenvFactory) -> None:
    with poetry_venv_factory("example_project") as virtual_env:
        issue_report = f"{uuid.uuid4()}.json"
        result = virtual_env.run(f"deptry . -o {issue_report}")

        expected_output = [
            "Scanning 2 files...",
            "",
            stylize(
                "{BOLD}{file}{RESET}{CYAN}:{RESET} {BOLD}{RED}DEP002{RESET} 'isort' defined as a dependency but not"
                " used in the codebase",
                file=Path("pyproject.toml"),
            ),
            stylize(
                "{BOLD}{file}{RESET}{CYAN}:{RESET} {BOLD}{RED}DEP002{RESET} 'requests' defined as a dependency but"
                " not used in the codebase",
                file=Path("pyproject.toml"),
            ),
            stylize(
                "{BOLD}{file}{RESET}{CYAN}:{RESET}4{CYAN}:{RESET}8{CYAN}:{RESET} {BOLD}{RED}DEP004{RESET} 'black'"
                " imported but declared as a dev dependency",
                file=Path("src/main.py"),
            ),
            stylize(
                "{BOLD}{file}{RESET}{CYAN}:{RESET}6{CYAN}:{RESET}8{CYAN}:{RESET} {BOLD}{RED}DEP001{RESET} 'white'"
                " imported but missing from the dependency definitions",
                file=Path("src/main.py"),
            ),
            stylize("{BOLD}{RED}Found 4 dependency issues.{RESET}"),
            "",
            "For more information, see the documentation: https://fpgmaas.github.io/deptry/",
            "",
        ]

        # Assert that we still write to console when generating a JSON report.
        assert result.stderr == "\n".join(expected_output)
        assert get_issues_report(Path(issue_report)) == [
            {
                "error": {
                    "code": "DEP002",
                    "message": "'isort' defined as a dependency but not used in the codebase",
                },
                "module": "isort",
                "location": {
                    "file": str(Path("pyproject.toml")),
                    "line": None,
                    "column": None,
                },
            },
            {
                "error": {
                    "code": "DEP002",
                    "message": "'requests' defined as a dependency but not used in the codebase",
                },
                "module": "requests",
                "location": {
                    "file": str(Path("pyproject.toml")),
                    "line": None,
                    "column": None,
                },
            },
            {
                "error": {
                    "code": "DEP004",
                    "message": "'black' imported but declared as a dev dependency",
                },
                "module": "black",
                "location": {
                    "file": str(Path("src/main.py")),
                    "line": 4,
                    "column": 8,
                },
            },
            {
                "error": {
                    "code": "DEP001",
                    "message": "'white' imported but missing from the dependency definitions",
                },
                "module": "white",
                "location": {
                    "file": str(Path("src/main.py")),
                    "line": 6,
                    "column": 8,
                },
            },
        ]


def test_cli_help() -> None:
    result = CliRunner().invoke(deptry, "--help")

    assert result.exit_code == 0



from __future__ import annotations

import json
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

from deptry.reporters.base import Reporter

if TYPE_CHECKING:
    from typing import Any


@dataclass
class JSONReporter(Reporter):
    json_output: str

    def report(self) -> None:
        output: list[dict[str, str | dict[str, Any]]] = []

        for violation in self.violations:
            output.append(
                {
                    "error": {
                        "code": violation.error_code,
                        "message": violation.get_error_message(),
                    },
                    "module": violation.issue.name,
                    "location": {
                        "file": str(violation.location.file),
                        "line": violation.location.line,
                        "column": violation.location.column,
                    },
                },
            )

        with Path(self.json_output).open("w", encoding="utf-8") as f:
            json.dump(output, f, ensure_ascii=False, indent=4)



from __future__ import annotations

from importlib.metadata import PackageNotFoundError
from pathlib import Path
from unittest.mock import patch

from deptry.dependency import Dependency


def test_simple_dependency() -> None:
    dependency = Dependency("click", Path("pyproject.toml"))
    assert dependency.name == "click"
    assert dependency.definition_file == Path("pyproject.toml")
    assert dependency.top_levels == {"click"}


def test_create_default_top_level_if_metadata_not_found() -> None:
    dependency = Dependency("Foo-bar", Path("foo/requirements.txt"))
    assert dependency.name == "Foo-bar"
    assert dependency.definition_file == Path("foo/requirements.txt")
    assert dependency.top_levels == {"foo_bar"}


def test_read_top_level_from_top_level_txt() -> None:
    """
    Read the top-levels.txt file
    """

    class MockDistribution:
        def __init__(self) -> None:
            pass

        def read_text(self, file_name: str) -> str:
            return "foo\nbar"

    with patch("deptry.dependency.metadata.distribution") as mock:
        mock.return_value = MockDistribution()
        dependency = Dependency("Foo-bar", Path("pyproject.toml"))

    assert dependency.name == "Foo-bar"
    assert dependency.definition_file == Path("pyproject.toml")
    assert dependency.top_levels == {"foo", "bar"}


def test_read_top_level_from_record() -> None:
    """
    Verify that if top-level.txt not found, an attempt is made to extract top-level module names from
    the metadata RECORD file.
    """

    class MockDistribution:
        def __init__(self) -> None:
            pass

        def read_text(self, file_name: str) -> str | None:
            if file_name == "RECORD":
                return """\
../../../bin/black,sha256=<HASH>,247
__pycache__/_black_version.cpython-311.pyc,,
_black_version.py,sha256=<HASH>,19
black/trans.cpython-39-darwin.so,sha256=<HASH>
black/trans.py,sha256=<HASH>
blackd/__init__.py,sha256=<HASH>
blackd/__main__.py,sha256=<HASH>
                """
            return None

    with patch("deptry.dependency.metadata.distribution") as mock:
        mock.return_value = MockDistribution()
        dependency = Dependency("Foo-bar", Path("pyproject.toml"))

    assert dependency.name == "Foo-bar"
    assert dependency.definition_file == Path("pyproject.toml")
    assert dependency.top_levels == {"_black_version", "black", "blackd"}


def test_read_top_level_from_predefined() -> None:
    """
    Verify that if there are predefined top-level module names it takes
    precedence over other lookup methods.
    """
    with patch("deptry.dependency.metadata.distribution") as mock:
        dependency = Dependency("Foo-bar", Path("pyproject.toml"), module_names=["foo"])

    assert dependency.name == "Foo-bar"
    assert dependency.definition_file == Path("pyproject.toml")
    assert dependency.top_levels == {"foo"}
    mock.return_value.read_text.assert_not_called()


def test_not_predefined_and_not_installed() -> None:
    """
    Use the fallback option of translating the package name.
    """

    with patch("deptry.dependency.metadata.distribution") as mock:
        mock.side_effect = PackageNotFoundError
        dependency = Dependency("Foo-bar", Path("pyproject.toml"))

    assert dependency.name == "Foo-bar"
    assert dependency.top_levels == {"foo_bar"}



from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING

import click
import pytest
from click import Argument

from deptry.config import read_configuration_from_pyproject_toml
from deptry.exceptions import InvalidPyprojectTOMLOptionsError
from tests.utils import run_within_dir

if TYPE_CHECKING:
    from _pytest.logging import LogCaptureFixture


click_command = click.Command(
    "",
    params=[
        Argument(param_decls=["exclude"]),
        Argument(param_decls=["extend_exclude"]),
        Argument(param_decls=["per_rule_ignores"]),
        Argument(param_decls=["ignore"]),
        Argument(param_decls=["ignore_notebooks"]),
        Argument(param_decls=["requirements_txt"]),
        Argument(param_decls=["requirements_txt_dev"]),
    ],
)


def test_read_configuration_from_pyproject_toml_exists(tmp_path: Path) -> None:
    click_context = click.Context(
        click_command,
        default_map={
            "exclude": ["bar"],
            "extend_exclude": ["foo"],
            "per_rule_ignores": {
                "DEP002": ["baz", "bar"],
            },
            "ignore": [],
            "requirements_txt": "requirements.txt",
            "requirements_txt_dev": ["requirements-dev.txt"],
        },
    )

    pyproject_toml_content = """
        [tool.deptry]
        exclude = ["foo", "bar"]
        extend_exclude = ["bar", "foo"]
        ignore_notebooks = true
        ignore = ["DEP001", "DEP002", "DEP003", "DEP004"]
        requirements_txt = "foo.txt"
        requirements_txt_dev = ["dev.txt", "tests.txt"]

        [tool.deptry.per_rule_ignores]
        DEP001 = ["baz", "foobar"]
        DEP002 = ["foo"]
        DEP003 = ["foobaz"]
        DEP004 = ["barfoo"]
    """

    with run_within_dir(tmp_path):
        pyproject_toml_path = Path("pyproject.toml")

        with pyproject_toml_path.open("w") as f:
            f.write(pyproject_toml_content)

        assert (
            read_configuration_from_pyproject_toml(click_context, click.UNPROCESSED(None), pyproject_toml_path)
            == pyproject_toml_path
        )

    assert click_context.default_map == {
        "exclude": ["foo", "bar"],
        "extend_exclude": ["bar", "foo"],
        "ignore_notebooks": True,
        "per_rule_ignores": {
            "DEP001": ["baz", "foobar"],
            "DEP002": ["foo"],
            "DEP003": ["foobaz"],
            "DEP004": ["barfoo"],
        },
        "ignore": ["DEP001", "DEP002", "DEP003", "DEP004"],
        "requirements_txt": "foo.txt",
        "requirements_txt_dev": ["dev.txt", "tests.txt"],
    }


def test_read_configuration_from_pyproject_toml_file_not_found(caplog: LogCaptureFixture) -> None:
    pyproject_toml_path = Path("a_non_existent_pyproject.toml")

    with caplog.at_level(logging.DEBUG):
        assert (
            read_configuration_from_pyproject_toml(
                click.Context(click_command), click.UNPROCESSED(None), pyproject_toml_path
            )
            == pyproject_toml_path
        )

    assert "No pyproject.toml file to read configuration from." in caplog.text


def test_read_configuration_from_pyproject_toml_file_without_deptry_section(
    caplog: LogCaptureFixture, tmp_path: Path
) -> None:
    pyproject_toml_content = """
        [tool.something]
        exclude = ["foo", "bar"]
    """

    with run_within_dir(tmp_path):
        pyproject_toml_path = Path("pyproject.toml")

        with pyproject_toml_path.open("w") as f:
            f.write(pyproject_toml_content)

        with caplog.at_level(logging.DEBUG):
            assert read_configuration_from_pyproject_toml(
                click.Context(click_command), click.UNPROCESSED(None), pyproject_toml_path
            ) == Path("pyproject.toml")

    assert "No configuration for deptry was found in pyproject.toml." in caplog.text


def test_read_configuration_from_pyproject_toml_file_with_invalid_options(
    caplog: LogCaptureFixture, tmp_path: Path
) -> None:
    pyproject_toml_content = """
        [tool.deptry]
        exclude = ["foo", "bar"]
        invalid_option = "nope"
        another_invalid_option = "still nope"
        extend_exclude = ["bar", "foo"]
    """

    with run_within_dir(tmp_path):
        pyproject_toml_path = Path("pyproject.toml")

        with pyproject_toml_path.open("w") as f:
            f.write(pyproject_toml_content)

        with pytest.raises(InvalidPyprojectTOMLOptionsError):
            assert read_configuration_from_pyproject_toml(
                click.Context(click_command), click.UNPROCESSED(None), pyproject_toml_path
            ) == Path("pyproject.toml")
